#!/bin/sh
# SydB☮x: seccomp and landlock based application sandbox with support for namespaces
# data/syd.sh: Defines 'esyd' command, the multi functional SydB☮x helper.
#
# esyd is written in portable shell.
# It should work fine with POSIX sh, Bash and Zsh.
# If you spot a problem running this with either of them,
# please report a bug at: https://todo.sr.ht/~alip/syd
#
# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
# SPDX-License-Identifier: GPL-3.0-or-later

esyd() {
    local cmd="${1}"

    shift
    case "${cmd}" in
    api)
        echo -n 3
        ;;
    check)
        syd --check
        local r=$?
        if [ -t 1 ]; then
            case $r in
            0)
                echo >&2 "$(tput setaf 10)Running under SydB☮x.$(tput sgr0)"
                ;;
            *)
                echo >&2 "$(tput setaf 9)Not running under SydB☮x.$(tput sgr0)"
                ;;
            esac
        fi
        return $r
        ;;
    stat)
        [ -c /dev/syd/stat ]
        ;;
    lock)
        [ -c '/dev/syd/lock:on' ]
        ;;
    unlock)
        [ -c '/dev/syd/lock:off' ]
        ;;
    exec_lock)
        [ -c '/dev/syd/lock:exec' ]
        ;;
    info)
        if [ -c /dev/syd ]; then
            if command -v jq >/dev/null 2>&1; then
                local out
                out=$(mktemp)
                if [ -n "$BASH_VERSION" ]; then
                    # We're in Bash
                    # This works with lock:exec
                    IFS=$'\n' readarray -t syd < /dev/syd
                    echo "${syd[*]}" > "${out}"
                elif [ -n "$ZSH_VERSION" ]; then
                    syd=$(</dev/syd)
                    echo "${syd}" > "${out}"
                else
                    # Fallback for POSIX sh
                    # This needs lock:off, doesn't work with lock:exec
                    cat /dev/syd > "${out}"
                fi

                if [ -t 1 ]; then
                    jq "${@}" < "${out}" | ${PAGER:-less}
                else
                    jq "${@}" < "${out}"
                fi

                r=$?
                rm -f "${out}"
                return $r
            else
                if [ -n "$BASH_VERSION" ]; then
                    # We're in Bash
                    # This works with lock:exec
                    IFS=$'\n' readarray -t syd < /dev/syd
                    echo "${syd[*]}"
                elif [ -n "$ZSH_VERSION" ]; then
                    syd=$(</dev/syd)
                    echo "${syd}"
                else
                    # Fallback for POSIX sh
                    # This needs lock:off, doesn't work with lock:exec
                    cat /dev/syd
                fi
            fi
        else
            return 1
        fi
        ;;
    exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        [ -c "$(syd exec ${@})" ]
        ;;
    kill)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'exec/kill' '+' "${@}"
        ;;
    enabled|enabled_path)
        [ -c '/dev/syd/sandbox/write?' ]
        ;;
    enable|enable_path)
        [ -c '/dev/syd/sandbox/write:on' ]
        ;;
    disable|disable_path)
        [ -c '/dev/syd/sandbox/write:off' ]
        ;;
    enabled_read)
        [ -c '/dev/syd/sandbox/read?' ]
        ;;
    enable_read)
        [ -c '/dev/syd/sandbox/read:on' ]
        ;;
    disable_read)
        [ -c '/dev/syd/sandbox/read:off' ]
        ;;
    enabled_stat)
        [ -c '/dev/syd/sandbox/stat?' ]
        ;;
    enable_stat)
        [ -c '/dev/syd/sandbox/stat:on' ]
        ;;
    disable_stat)
        [ -c '/dev/syd/sandbox/stat:off' ]
        ;;
    enabled_exec)
        [ -c '/dev/syd/sandbox/exec?' ]
        ;;
    enable_exec)
        [ -c '/dev/syd/sandbox/exec:on' ]
        ;;
    disable_exec)
        [ -c '/dev/syd/sandbox/exec:off' ]
        ;;
    enabled_net)
        [ -c '/dev/syd/sandbox/net?' ]
        ;;
    enable_net)
        [ -c '/dev/syd/sandbox/net:on' ]
        ;;
    disable_net)
        [ -c '/dev/syd/sandbox/net:off' ]
        ;;
    allow|allow_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/write' '+' "${@}"
        ;;
    disallow|disallow_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/write' '-' "${@}"
        ;;
    deny|deny_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/write' '+' "${@}"
        ;;
    nodeny|nodeny_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/write' '-' "${@}"
        ;;
    allow_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/read' '+' "${@}"
        ;;
    disallow_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/read' '-' "${@}"
        ;;
    deny_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/read' '+' "${@}"
        ;;
    nodeny_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/read' '-' "${@}"
        ;;
    allow_stat)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/stat' '+' "${@}"
        ;;
    disallow_stat)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/stat' '-' "${@}"
        ;;
    deny_stat)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/stat' '+' "${@}"
        ;;
    nodeny_stat)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/stat' '-' "${@}"
        ;;
    allow_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/exec' '+' "${@}"
        ;;
    disallow_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'allowlist/exec' '-' "${@}"
        ;;
    deny_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/exec' '+' "${@}"
        ;;
    nodeny_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'denylist/exec' '-' "${@}"
        ;;
    allow_net)
        local c="allowlist/net/bind"
        [ "${1}" == "--connect" ] && c="allowlist/net/connect" && shift
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net "${c}" '+' "${@}"
        ;;
    disallow_net)
        local c="allowlist/net/bind"
        [ "${1}" == "--connect" ] && c="allowlist/net/connect" && shift
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net "${c}" '-' "${@}"
        ;;
    deny_net)
        local c="denylist/net/bind"
        [ "${1}" == "--connect" ] && c="denylist/net/connect" && shift
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net "${c}" '+' "${@}"
        ;;
    nodeny_net)
        local c="denylist/net/bind"
        [ "${1}" == "--connect" ] && c="denylist/net/connect" && shift
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net "${c}" '-' "${@}"
        ;;
    addfilter|addfilter_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/write' '+' "${@}"
        ;;
    rmfilter|rmfilter_path)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/write' '-' "${@}"
        ;;
    addfilter_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/read' '+' "${@}"
        ;;
    rmfilter_read)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/read' '-' "${@}"
        ;;
    addfilter_stat)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/stat' '+' "${@}"
        ;;
    rmfilter_stat)
        if  [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/stat' '-' "${@}"
        ;;
    addfilter_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/exec' '+' "${@}"
        ;;
    rmfilter_exec)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_path 'filter/exec' '-' "${@}"
        ;;
    addfilter_net)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net 'filter/net' '+' "${@}"
        ;;
    rmfilter_net)
        if [ ${#} -lt 1 ]; then
            echo >&2 "$(tput setaf 9)esyd: ${cmd} takes at least one extra argument$(tput sgr0)"
            return 1
        fi
        _esyd_net 'filter/net' '-' "${@}"
        ;;
    help|'')
        if [ -t 1 ]; then
            esyd help | ${PAGER:-less}
            return $?
        fi

        local cblu ccya cgre cmag cred cres cwhi cyel
        cblu=$(tput setaf 12)
        ccya=$(tput setaf 14)
        cgre=$(tput setaf 10)
        cmag=$(tput setaf 13)
        cred=$(tput setaf 9)
        cres=$(printf '\033[0m')
        cwhi=$(tput setaf 15)
        cyel=$(tput setaf 11)
        cat <<EOF
${cred}esyd${cres} -- ${cyel}multi functional${cres} ${cred}SydB☮x${cres} ${cyel}helper${cres}
${cmag}Copyright (c) 2023 Ali Polatel <alip@chesswob.org>${cres}
${cmag}SPDX-License-Identifier: GPL-3.0-or-later${cres}

${cred}# Subcommands${cres}
${cgre}api${cres}
    ${cwhi}Print SydB☮x API version number${cres}
${cgre}check${cres}
    ${cwhi}Return true if running under SydB☮x${cres}
${cgre}lock${cres}
    ${cwhi}Lock SydB☮x, esyd commands will no longer work${cres}
${cgre}exec_lock${cres}
    ${cwhi}Lock SydB☮x for all processes but the SydB☮x exec child${cres}
${cgre}unlock${cres}
    ${cwhi}Unlock the SydB☮x, make it available to all processes rather than just the SydB☮x exec child${cres}
${cgre}info${cres} ${ccya}jq-args...${cres}
    ${cwhi}Print SydB☮x sandbox state as JSON on standard output${cres}
    ${cwhi}If "jq" is in PATH, pass the arguments and pipe the output to jq${cres}
${cgre}stat${cres}
    ${cwhi}Print SydB☮x sandbox state on standard error${cres}
${cgre}exec ${ccya}cmd${cres} ${cyel}args..${cres}
    ${cwhi}Execute a command outside the sandbox without sandboxing${cres}
${cgre}kill${cres} ${ccya}glob${cres}
    ${cwhi}Kill any attempt to execute a path matching the given glob pattern${cres}
${cgre}enabled${cres}, ${cgre}enabled_path${cres}
    ${cwhi}Return true if write sandboxing is enabled${cres}
${cgre}enable${cres}, ${cgre}enable_path${cres}
    ${cwhi}Enable write sandboxing${cres}
${cgre}disable${cres}, ${cgre}disable_path${cres}
    ${cwhi}Disable write sandboxing${cres}
${cgre}enabled_read${cres}
    ${cwhi}Return true if read sandboxing is enabled${cres}
${cgre}enable_read${cres}
    ${cwhi}Enable read sandboxing${cres}
${cgre}disable_read${cres}
    ${cwhi}Disable read sandboxing${cres}
${cgre}enabled_stat${cres}
    ${cwhi}Return true if stat sandboxing is enabled${cres}
${cgre}enable_stat${cres}
    ${cwhi}Enable stat sandboxing${cres}
${cgre}disable_stat${cres}
    ${cwhi}Disable stat sandboxing${cres}
${cgre}enabled_exec${cres}
    ${cwhi}Return true if exec sandboxing is enabled${cres}
${cgre}enable_exec${cres}
    ${cwhi}Enable exec sandboxing${cres}
${cgre}disable_exec${cres}
    ${cwhi}Disable exec sandboxing${cres}
${cgre}enabled_net${cres}
    ${cwhi}Return true if network sandboxing is enabled${cres}
${cgre}enable_net${cres}
    ${cwhi}Enable network sandboxing${cres}
${cgre}disable_net${cres}
    ${cwhi}Disable network sandboxing${cres}
${cgre}allow${cres}, ${cgre}allow_path${cres} ${ccya}glob${cres}
    ${cwhi}Allowlist the given glob pattern for write sandboxing${cres}
${cgre}disallow${cres}, ${cgre}disallow_path${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the allowlist for write sandboxing${cres}
${cgre}deny${cres}, ${cgre}deny_path${cres} ${ccya}glob${cres}
    ${cwhi}Denylist the given glob pattern for write sandboxing${cres}
${cgre}nodeny${cres}, ${cgre}nodeny_path ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the denylist for write sandboxing${cres}
${cgre}allow_read${cres} ${ccya}glob${cres}
    ${cwhi}Allowlist the given glob pattern for read sandboxing${cres}
${cgre}disallow_read${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the allowlist for read sandboxing${cres}
${cgre}deny_read${cres} ${ccya}glob${cres}
    ${cwhi}Denylist the given glob pattern for read sandboxing${cres}
${cgre}nodeny_read${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattenr from the denylist for read sandboxing${cres}
${cgre}allow_stat${cres} ${ccya}glob${cres}
    ${cwhi}Allowlist the given glob pattern for stat sandboxing${cres}
${cgre}disallow_stat${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the allowlist for stat sandboxing${cres}
${cgre}deny_stat${cres} ${ccya}glob${cres}
    ${cwhi}Denylist the given glob pattern for stat sandboxing${cres}
${cgre}nodeny_stat${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattenr from the denylist for stat sandboxing${cres}
${cgre}allow_exec${cres} ${ccya}glob${cres}
    ${cwhi}Allowlist the given glob pattern for exec sandboxing${cres}
${cgre}disallow_exec${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the allowlist for exec sandboxing${cres}
${cgre}deny_exec${cres} ${ccya}glob${cres}
    ${cwhi}Denylist the given glob pattern for exec sandboxing${cres}
${cgre}nodeny_exec${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the denylist for exec sandboxing${cres}
${cgre}allow_net${cres} ${cblu}[--connect]${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Allowlist the given network address for network bind or connect sandboxing${cres}
${cgre}disallow_net${cres} ${cblu}[--connect]${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Removes the given network address (Ipv4,6), or the glob pattern (UNIX domain sockets)${cres}
    ${cwhi}from the allowlist for network bind or connect sandboxing${cres}
${cgre}deny_net${cres} ${cblu}[--connect]${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Denylist the given network address (Ipv4,6) or the glob pattern (UNIX domain sockets)${cres}
    ${cwhi}for network bind or connect sandboxing${cres}
${cgre}nodeny_net${cres} ${cblu}[--connect]${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Removes the given network address (Ipv4,6) or the glob pattern (UNIX domain sockets)${cres}
    ${cwhi}from the denylist for network bind or connect sandboxing${cres}
${cgre}addfilter${cres}, ${cgre}addfilter_path${cres} ${ccya}glob${cres}
    ${cwhi}Adds the given glob pattern to the list of access violation filters for write sandboxing${cres}
${cgre}rmfilter${cres}, ${cgre}rmfilter_path${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the list of access violation filters for write sandboxing${cres}
${cgre}addfilter_read${cres} ${ccya}glob${cres}
    ${cwhi}Adds the given glob pattern to the list of access violation filters for read sandboxing${cres}
${cgre}rmfilter_read${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the list of access violation filters for read sandboxing${cres}
${cgre}addfilter_stat${cres} ${ccya}glob${cres}
    ${cwhi}Adds the given glob pattern to the list of access violation filters for stat sandboxing${cres}
${cgre}rmfilter_stat${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the list of access violation filters for stat sandboxing${cres}
${cgre}addfilter_exec${cres} ${ccya}glob${cres}
    ${cwhi}Adds the given glob pattern to the list of access violation filters for exec sandboxing${cres}
${cgre}rmfilter_exec${cres} ${ccya}glob${cres}
    ${cwhi}Removes the given glob pattern from the list of access violation filters for exec sandboxing${cres}
${cgre}addfilter_net${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Adds the network address (Ipv4,6) or the glob pattern (UNIX domain sockets)${cres}
    ${cwhi}to the list of access violation filters for network sandboxing${cres}
${cgre}rmfilter_net${cres} ${ccya}glob${cres}|${cyel}cidr@port[-port]${cres}
    ${cwhi}Removes the network address (Ipv4,6) or the glob pattern (UNIX domain sockets)${cres}
    ${cwhi}from the list of access violation filters for network sandboxing${cres}
EOF
        ;;
    *)
        local cmd="$(tput setaf 10)${cmd}$(printf '\033[0m')"
        local hlp="$(tput setaf 13)esyd help$(printf '\033[0m')"
        echo >&2 "$(tput setaf 9)esyd: subcommand$(printf '\033[0m') ${cmd} $(tput setaf 9)unrecognised, use ${hlp} $(tput setaf 9)for a list of supported subcommands.$(printf '\033[0m')"
        return 1
        ;;
    esac
}

_esyd_path()
{
    local cmd="${1}"
    local op="${2}"

    case "${op}" in
    '+'|'-')
        ;;
    *)
        echo >&2 "$(tput setaf 9)esyd_path: invalid operation character '${op}'$(printf '\033[0m')"
        return 1
        ;;
    esac

    shift 2

    local ret=0
    local path
    for path in "${@}"; do
        case "${path}" in
        /*)
            [ -c /dev/syd/"${cmd}${op}${path}" ] || ret=$?
            ;;
        *)
            echo >&2 "$(tput setaf 9)esyd_path: expects absolute path, got: ${path}$(printf '\033[0m')"
            return 1
            ;;
        esac
    done
    return $ret
}

_esyd_net()
{
    local cmd="${1}"
    local op="${2}"

    case "${op}" in
    '+'|'-')
        ;;
    *)
        echo >&2 "$(tput setaf 9)esyd_net: invalid operation character '${op}'$(printf '\033[0m')"
        return 1
        ;;
    esac

    shift 2

    local ret=0
    while [ ${#} -gt 0 ] ; do
        # SydB☮x does input validation so we don't do any here.
        [ -c "/dev/syd/${cmd}${op}${1}" ] || ret=$?
        shift
    done
    return $ret
}
