//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/lib.rs: Common utility functions
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! # SydB☮x: The ☮ther SⒶndbøx
//!
//! [![Shine On You Crazy Diamond!](https://img.shields.io/badge/Shine%20On%20You%20Crazy%20Diamond!-8A2BE2)](https://en.wikipedia.org/wiki/Syd_Barrett)
//! [![license](https://img.shields.io/crates/l/jja.svg)](https://git.sr.ht/~alip/syd/tree/main/item/COPYING)
//! [![msrv](https://img.shields.io/badge/rustc-1.70%2B-green?style=plastic)](https://blog.rust-lang.org/2023/06/01/Rust-1.70.0.html)
//! [![build status](https://builds.sr.ht/~alip/syd.svg)](https://builds.sr.ht/~alip/syd?)
//! [![maintenance-status](https://img.shields.io/badge/maintenance-actively--developed-brightgreen.svg)](https://git.sr.ht/~alip/syd)
//! [![dependency status](https://deps.rs/repo/sourcehut/~alip/syd/status.svg)](https://deps.rs/repo/sourcehut/~alip/syd)
//! [![repology](https://repology.org/badge/latest-versions/sydbox.svg)](https://repology.org/project/sydbox/versions)
//!
//! [![SydB☮x](https://git.sr.ht/~alip/syd/blob/main/data/syd.png)](https://todo.sr.ht/~alip/syd)
//! [![GNU](https://web.archive.org/web/20221222061733if_/https://dev.exherbo.org/~alip/images/gnu.png)](https://www.gnu.org/philosophy/philosophy.html)
//! [![Linux](https://chesswob.org/jja/tux.png)](https://www.kernel.org/category/about.html)
//! [![Exherbo](https://web.archive.org/web/20230518155203if_/https://dev.exherbo.org/~alip/images/zebrapig.png)](https://www.exherbolinux.org/docs/gettingstarted.html)
//! [![musl libc](https://www.chesswob.org/jja/musl-inside.png)](https://www.musl-libc.org/)
//! [![libsecc☮mp](https://web.archive.org/web/20221222061720if_/https://dev.exherbo.org/~alip/images/libseccomp.png)](https://github.com/seccomp/libseccomp)
//! [![Paludis](http://paludis.exherbolinux.org/paludis_270.png)](https://paludis.exherbolinux.org)
//!
//! SydB☮x is a **seccomp**(2) based sandboxing utility for modern Linux\[\>=5.6\]
//! machines to sandbox unwanted process access to filesystem and network resources.
//! SydB☮x requires *no root access* and *no ptrace* rights. All you need is a
//! recent Linux kernel and libsecc☮mp which is available on many different
//! architectures, including **x86**, **x86\_64**, **x32**, **arm**, **aarch64**,
//! **mips**, **mips64**... This makes it very easy for a regular user to use. This is
//! the motto of SydB☮x: *bring easy, simple, flexible and powerful access restriction
//! to the Linux user!*
//!
//! The basic idea of SydB☮x is to run a command under certain restrictions. These
//! restrictions define which system calls the command is permitted to run and which
//! argument values are permitted for the given system call. The restrictions may be
//! applied via two ways.  *seccomp-bpf* can be used to apply simple Secure Computing
//! user filters to run sandboxing fully on kernel space, and *seccomp-notify*
//! functionality can be used to run sandboxing on kernel space and fallback to user
//! space to dereference pointer arguments of system calls (**See
//! [Security](#security) about `TOCTOU` et. al**), which are one of
//! **[pathname](https://en.wikipedia.org/wiki/Path_(computing))**, **[UNIX socket
//! address](https://en.wikipedia.org/wiki/Unix_domain_socket)**,
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** or
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** network address, and make dynamic
//! decisions using [Unix shell style patterns](https://docs.rs/globset) such as
//! `allowlist/write+/home/sydbox/***`, or `allowlist/write+/run/user/*/pulse` for
//! **[pathnames](https://en.wikipedia.org/wiki/Path_(computing))**, and using
//! **[CIDR](https://docs.rs/ipnetwork)** notation such as
//! `allowlist/net/connect+127.0.0.1/8@9050`, or
//! `allowlist/net/connect+::1/8@9050` for
//! **[IPv4](https://en.wikipedia.org/wiki/IPv4)** and
//! **[IPv6](https://en.wikipedia.org/wiki/IPv6)** addresses and perform an action
//! which is by default denying the system call with an appropriate error, which is
//! usually **access denied**, aka `EACCES`. For default disallowed system calls,
//! such as `ptrace` or `process_vm_writev` (**See [Security](#security) about
//! `TOCTOU` et. al**) SydB☮x returns `EACCES` as well.
//!
//! To be able to use SydB☮x, you need a recent Linux kernel with the system calls
//! **pidfd_getfd**, **pidfd_send_signal**. The Secure Computing facility of the
//! Linux kernel should support the **SECCOMP_USER_NOTIF_FLAG_CONTINUE** operation.
//! It is recommended to have the **CONFIG_CROSS_MEMORY_ATTACH** kernel option
//! enabled, if this option is not enabled, sydbox will fallback to reading/writing
//! from `/proc/$pid/mem`. Linux-5.11 or later is recommended.

// We like clean and simple code with documentation.
// Keep in sync with main.rs.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
//#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

/// Compatibility code for different libcs
pub(crate) mod compat;
/// Static configuration, edit & recompile!
pub mod config;
/// Filesystem utilities
pub mod fs;
/// Secure computing hooks
#[allow(clippy::as_ptr_cast_mut)]
#[allow(clippy::cast_sign_loss)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod hook;
/// Simple logging on standard error using JSON lines
pub mod log;
/// /proc utilities
pub mod proc;
/// Sandbox configuration
pub mod sandbox;

// Vendored crates:
/// Interface to LandLock LSM
#[allow(dead_code)]
#[allow(missing_docs)]
#[allow(clippy::as_underscore)]
#[allow(clippy::cast_possible_truncation)]
#[allow(clippy::decimal_literal_representation)]
#[allow(clippy::missing_safety_doc)]
#[allow(clippy::type_complexity)]
#[allow(clippy::undocumented_unsafe_blocks)]
pub mod landlock;
/// The low-level interface for linux namespaces (containers)
pub mod unshare;

use std::{
    ffi::OsString,
    fs::File,
    io::{BufRead, BufReader},
    os::unix::ffi::{OsStrExt, OsStringExt},
    path::{Path, PathBuf},
    process::exit,
};

use libseccomp::{ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
use nix::{
    errno::Errno,
    fcntl::OFlag,
    sys::{
        signal::{sigaction, SaFlags, SigAction, SigHandler, Signal},
        stat::{stat, Mode},
        wait::{waitpid, WaitStatus},
    },
    unistd::{fork, ForkResult, Pid, Uid, User},
};
use once_cell::sync::Lazy;

/* Utilities */

/// Returns true if we are running under SydB☮x.
#[allow(clippy::disallowed_methods)]
pub fn syd_enabled() -> bool {
    // This will not work if the sandbox is locked.
    // Path::new("/dev/syd").exists() || Path::new("/dev/sydbox").exists()
    // SAFETY: In libc, we trust.
    match unsafe { fork() } {
        Ok(ForkResult::Parent { child, .. }) => {
            match waitpid(child, None) {
                Ok(WaitStatus::Exited(_, code)) => {
                    // Check the child's exit status.
                    // Exit status of 0 means SydB☮x is enabled.
                    code == 0
                }
                _ => {
                    // If there's an error waiting on the
                    // child, assume SydB☮x is not enabled.
                    false
                }
            }
        }
        Ok(ForkResult::Child) => {
            let mut ctx = match ScmpFilterContext::new_filter(ScmpAction::Allow) {
                Ok(ctx) => ctx,
                Err(_) => exit(1),
            };

            let syscall = ScmpSyscall::new("open");
            if ctx.add_rule(ScmpAction::Notify, syscall).is_err() {
                exit(1);
            }

            if ctx.load().is_err() && Errno::last() == Errno::EBUSY {
                // seccomp filter exists
                // SydB☮x is in business.
                exit(0);
            } else {
                // seccomp filter does not exist
                exit(1);
            }
        }
        Err(_) => {
            // If there's an error forking,
            // assume SydB☮x is not enabled.
            false
        }
    }
}

/// Get the umask of the given `Pid` or current process'.
pub(crate) fn proc_umask(pid: Option<Pid>) -> Result<Mode, Errno> {
    // Construct path to the appropriate status file
    let path = match pid {
        Some(pid_value) => Path::new("/proc")
            .join(pid_value.to_string())
            .join("status"),
        None => Path::new("/proc/self/status").to_path_buf(),
    };

    // Open the file
    let file = File::open(path)
        .map_err(|error| Errno::from_i32(error.raw_os_error().unwrap_or(nix::libc::ENOENT)))?;
    let reader = BufReader::new(file);

    // Search for Umask line
    for line in reader
        .lines()
        .flatten()
        .filter(|line| line.starts_with("Umask:"))
    {
        // Parse the umask value
        let parts: Vec<&str> = line.split_whitespace().collect();
        if parts.len() == 2 {
            if let Ok(umask) = u32::from_str_radix(parts[1], 8) {
                return Ok(Mode::from_bits_truncate(umask));
            } else {
                return Err(Errno::EINVAL);
            }
        }
    }

    Err(Errno::ENOENT)
}

/// Returns the name of the libsecc☮mp native architecture.
pub fn seccomp_arch_native_name() -> Option<&'static str> {
    match ScmpArch::native() {
        ScmpArch::X86 => Some("x86"),
        ScmpArch::X8664 => Some("x86_64"),
        ScmpArch::X32 => Some("x32"),
        ScmpArch::Arm => Some("arm"),
        ScmpArch::Aarch64 => Some("aarch64"),
        ScmpArch::Mips => Some("mips"),
        ScmpArch::Mips64 => Some("mips64"),
        ScmpArch::Mips64N32 => Some("mips64n32"),
        ScmpArch::Mipsel => Some("mipsel"),
        ScmpArch::Mipsel64 => Some("mipsel64"),
        ScmpArch::Mipsel64N32 => Some("mipsel64n32"),
        ScmpArch::Ppc => Some("ppc"),
        ScmpArch::Ppc64 => Some("ppc64"),
        ScmpArch::Ppc64Le => Some("ppc64le"),
        ScmpArch::S390 => Some("s390"),
        ScmpArch::S390X => Some("s390s"),
        ScmpArch::Parisc => Some("parisc"),
        ScmpArch::Parisc64 => Some("parisc64"),
        ScmpArch::Riscv64 => Some("riscv64"),
        _ => None,
    }
}

/// Given a `Uid`, return the user name of the user.
/// On any error conditions, return "nobody".
pub fn get_user_name(uid: Uid) -> String {
    match User::from_uid(uid) {
        Ok(Some(user)) => user.name,
        _ => "nobody".to_string(),
    }
}

/// Given a username, return the home directory of the user.
/// On any error conditions, return "/var/empty".
pub fn get_user_home(username: &str) -> String {
    // Fetch user details.
    match User::from_name(username) {
        Ok(Some(user)) => user.dir.to_string_lossy().to_string(),
        _ => "/var/empty".to_string(),
    }
}

// Sets the specified signal to be ignored.
//
// This function utilizes the `sigaction` system call to set the specified signal's action
// to `SIG_IGN`, effectively causing the process to ignore that signal.
//
// # Arguments
//
// * `signal` - The signal number (e.g., `SIGTSTP`, `SIGTTIN`, `SIGTTOU`).
//
// # Returns
//
// * `Result<(), Error>` - Returns `Ok(())` if successful, or an error if the operation fails.
//
// # Example
//
// ```no_run
// use nix::sys::signal::SIGTSTP;
//
// let result = syd::ignore_signal(SIGTSTP);
// assert!(result.is_ok());
// ```
pub(crate) fn ignore_signal(signal: Signal) -> Result<(), Errno> {
    let sig_action = SigAction::new(
        SigHandler::SigIgn, // Set to ignore
        SaFlags::empty(),
        nix::sys::signal::SigSet::empty(),
    );

    // SAFETY: The unsafe call to `sigaction` is used to set the signal's disposition
    // to "ignore". We're not invoking any handlers or performing any operations that
    // could lead to data races or other undefined behaviors. Hence, it's safe to call
    // in this context.
    unsafe {
        sigaction(signal, &sig_action)
            .map(|_| ())
            .map_err(|_| Errno::last())
    }
}

type BytesPair<'a> = (&'a [u8], Vec<u8>);
type BytesVec<'a> = Vec<BytesPair<'a>>;

/// Replace /proc/self with /proc/pid and /dev/fd with /proc/pid/fd
pub(crate) fn replace_self_with_pid<P: AsRef<Path>>(path: P, pid: Pid) -> PathBuf {
    let mut bytes = path.as_ref().as_os_str().as_bytes().to_vec();

    let replacements: BytesVec = vec![
        (b"/proc/self", format!("/proc/{pid}").into_bytes()),
        (b"/proc/thread-self", format!("/proc/{pid}").into_bytes()),
        (b"/dev/fd", format!("/proc/{pid}/fd").into_bytes()),
        (b"/dev/stdin", format!("/proc/{pid}/fd/0").into_bytes()),
        (b"/dev/stdout", format!("/proc/{pid}/fd/1").into_bytes()),
        (b"/dev/stderr", format!("/proc/{pid}/fd/2").into_bytes()),
    ];

    for (needle, replacement) in replacements {
        if let Some(pos) = bytes
            .windows(needle.len())
            .position(|window| window == needle)
        {
            // Ensure either it's the end of the string or followed by a '/'
            if pos.saturating_add(needle.len()) == bytes.len()
                || bytes[pos.saturating_add(needle.len())] == b'/'
            {
                bytes.splice(
                    pos..pos.saturating_add(needle.len()),
                    replacement.iter().copied(),
                );
            }
        }
    }

    PathBuf::from(OsString::from_vec(bytes))
}

/// Return system call priority by system call name.
#[inline(always)]
pub(crate) fn syscall_priority(name: &str) -> u8 {
    if name.starts_with("stat") {
        255
    } else if name.contains("open") {
        225
    } else if name.starts_with("getdents") {
        200
    } else if name.starts_with("access") {
        150
    } else if name.starts_with("execve") {
        100
    } else if name.starts_with("unlink") {
        75
    } else if name.starts_with("mkdir") {
        50
    } else {
        25
    }
}

/// Conerts an `OFlag` to a `Vec` of `String`s.
#[allow(clippy::cognitive_complexity)]
pub(crate) fn oflag_fmt(flags: OFlag) -> Vec<&'static str> {
    let mut vec = Vec::new();

    match flags & OFlag::O_ACCMODE {
        OFlag::O_RDONLY => {
            vec.push("O_RDONLY");
        }
        OFlag::O_WRONLY => {
            vec.push("O_WRONLY");
        }
        OFlag::O_RDWR => {
            vec.push("O_RDWR");
        }
        _ => {}
    };
    if flags.contains(OFlag::O_APPEND) {
        vec.push("O_APPEND");
    }
    if flags.contains(OFlag::O_ASYNC) {
        vec.push("O_ASYNC");
    }
    if flags.contains(OFlag::O_CLOEXEC) {
        vec.push("O_CLOEXEC");
    }
    if flags.contains(OFlag::O_CREAT) {
        vec.push("O_CREAT");
    }
    if flags.contains(OFlag::O_DIRECTORY) {
        vec.push("O_DIRECTORY");
    }
    if flags.contains(OFlag::O_EXCL) {
        vec.push("O_EXCL");
    }
    if flags.contains(OFlag::O_NOCTTY) {
        vec.push("O_NOCTTY");
    }
    if flags.contains(OFlag::O_NOFOLLOW) {
        vec.push("O_NOFOLLOW");
    }
    if flags.contains(OFlag::O_NONBLOCK) {
        vec.push("O_NONBLOCK");
    }
    if flags.contains(OFlag::O_SYNC) {
        vec.push("O_SYNC");
    }
    if flags.contains(OFlag::O_TRUNC) {
        vec.push("O_TRUNC");
    }

    vec
}

// device id and inode.
type FileId = (u64, u64);

pub(crate) static DEV_STAT: Lazy<Option<FileId>> = Lazy::new(|| get_stat("/dev"));

fn get_stat(path: &str) -> Option<FileId> {
    match stat(path) {
        Ok(metadata) => Some((metadata.st_dev, metadata.st_ino)),
        Err(_) => None,
    }
}

/// Like `std::path::Path::file_name()` but handles trailing slash transparently.
pub(crate) fn path_filename<P: AsRef<Path>>(path: P) -> PathBuf {
    let path = path.as_ref();
    let mut name = path
        .file_name()
        .map(PathBuf::from)
        .unwrap_or(PathBuf::from("."));

    // Handle trailing slash.
    if path.as_os_str().as_bytes().ends_with(b"/") {
        name.push("");
    }

    name
}

#[cfg(test)]
mod tests {
    use nix::{sys::stat::umask, unistd::Pid};

    use super::*;

    // FIXME: This test is racy and fails randomly.
    #[ignore]
    #[test]
    fn test_self_umask() {
        let my_umask = umask(Mode::empty());
        umask(my_umask); // Restore umask

        let result = proc_umask(None).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
    }

    #[test]
    fn test_invalid_pid() {
        let result = proc_umask(Some(Pid::from_raw(i32::MAX)));
        assert!(result.is_err(), "{result:?}");
    }

    /*
     * FIXME: This fails randomly, however the next test
     * passes consistently, WTF?
     */
    #[ignore]
    #[test]
    fn test_known_pid_umask() {
        // Testing against the current process.
        let my_umask = Mode::from_bits_truncate(0o0077);
        let old_umask = umask(my_umask);
        let result = proc_umask(None).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        let result = proc_umask(Some(Pid::this())).unwrap();
        assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        umask(old_umask); // Restore umask
    }

    #[test]
    fn test_parsing_valid_umask_values() {
        // This test sets various umask values and then checks if our function correctly identifies them.
        let umasks = [
            Mode::from_bits_truncate(0o0000),
            Mode::from_bits_truncate(0o0002),
            Mode::from_bits_truncate(0o0022),
            Mode::from_bits_truncate(0o0077),
            Mode::from_bits_truncate(0o0777),
        ];

        for &my_umask in &umasks {
            umask(my_umask);
            let result = proc_umask(None).unwrap();
            assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
            let result = proc_umask(Some(Pid::this())).unwrap();
            assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        }

        // Resetting the umask to a default value after test
        umask(Mode::from_bits_truncate(0o0022));
    }

    #[test]
    fn test_replace_self_with_pid_only() {
        let result = replace_self_with_pid("/proc/self/cmdline", Pid::from_raw(1234));
        assert_eq!(result, PathBuf::from("/proc/1234/cmdline"));
    }

    #[test]
    fn test_replace_fd_only() {
        let result = replace_self_with_pid("/dev/fd/3", Pid::from_raw(5678));
        assert_eq!(result, PathBuf::from("/proc/5678/fd/3"));
    }

    #[test]
    fn test_no_replacement() {
        let result = replace_self_with_pid("/some/random/path", Pid::from_raw(91011));
        assert_eq!(result, PathBuf::from("/some/random/path"));
    }

    #[test]
    fn test_replace_both_patterns() {
        let result = replace_self_with_pid("/proc/self/cmdline/dev/fd/3", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/cmdline/proc/1213/fd/3"));
    }

    #[test]
    fn test_replace_dev_stdin() {
        let result = replace_self_with_pid("/dev/stdin", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/0"));
    }

    #[test]
    fn test_replace_dev_stdout() {
        let result = replace_self_with_pid("/dev/stdout", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/1"));
    }

    #[test]
    fn test_replace_dev_stderr() {
        let result = replace_self_with_pid("/dev/stderr", Pid::from_raw(1213));
        assert_eq!(result, PathBuf::from("/proc/1213/fd/2"));
    }
}

///
/// TAROT
/// Check `esyd info .tarot`
pub(crate) fn draw() -> String {
    let pick = Tarot::new();

    if pick.reversed {
        format!("{} (Reversed) - {}", pick.card, pick.side)
    } else {
        format!("{} - {}", pick.card, pick.side)
    }
}

struct Tarot {
    card: &'static str,
    side: &'static str,
    reversed: bool,
}

impl Tarot {
    fn new() -> Self {
        let mut card: usize = 0;
        // SAFETY: In libc we trust.
        unsafe {
            nix::libc::getrandom(
                std::ptr::addr_of_mut!(card) as *mut nix::libc::c_void,
                std::mem::size_of_val(&card),
                0,
            );
            card %= TAROT_MAX;
        }

        let mut reversed_value = 0;
        // SAFETY: In libc we trust.
        unsafe {
            nix::libc::getrandom(
                std::ptr::addr_of_mut!(reversed_value) as *mut nix::libc::c_void,
                std::mem::size_of_val(&reversed_value),
                0,
            );
        }

        if reversed_value % 2 == 0 {
            Self {
                card: TAROT_CARDS[card],
                side: TAROT_SIDES[card].0,
                reversed: false,
            }
        } else {
            Self {
                card: TAROT_CARDS[card],
                side: TAROT_SIDES[card].1,
                reversed: true,
            }
        }
    }
}

const TAROT_MAX: usize = 78;
const TAROT_CARDS: [&str; TAROT_MAX] = [
    /* Major Arcana */
    "The Fool",
    "The Magician",
    "The High Priestess",
    "The Empress",
    "The Emperor",
    "The Heirophant",
    "The Lovers",
    "The Chariot",
    "Strength",
    "The Hermit",
    "Wheel of Fortune",
    "Justice",
    "The Hanged Man",
    "Death",
    "Temperance",
    "The Devil",
    "The Tower",
    "The Star",
    "The Moon",
    "The Sun",
    "Judgment",
    "The World",
    /* Cups */
    "The King of Cups",
    "The Queen of Cups",
    "The Knight of Cups",
    "The Page of Cups",
    "The Ten of Cups",
    "The Nine of Cups",
    "The Eight of Cups",
    "The Seven of Cups",
    "The Six of Cups",
    "The Five of Cups",
    "The Four of Cups",
    "The Three of Cups",
    "The Two of Cups",
    "The Ace of Cups",
    /* Swords */
    "The King of Swords",
    "The Queen of Swords",
    "The Knight of Swords",
    "The Page of Swords",
    "The Ten of Swords",
    "The Nine of Swords",
    "The Eight of Swords",
    "The Seven of Swords",
    "The Six of Swords",
    "The Five of Swords",
    "The Four of Swords",
    "The Three of Swords",
    "The Two of Swords",
    "The Ace of Swords",
    /* Wands */
    "The King of Wands",
    "The Queen of Wands",
    "The Knight of Wands",
    "The Page of Wands",
    "The Ten of Wands",
    "The Nine of Wands",
    "The Eight of Wands",
    "The Seven of Wands",
    "The Six of Wands",
    "The Five of Wands",
    "The Four of Wands",
    "The Three of Wands",
    "The Two of Wands",
    "The Ace of Wands",
    /* Pentacles */
    "The King of Pentacles",
    "The Queen of Pentacles",
    "The Knight of Pentacles",
    "The Page of Pentacles",
    "The Ten of Pentacles",
    "The Nine of Pentacles",
    "The Eight of Pentacles",
    "The Seven of Pentacles",
    "The Six of Pentacles",
    "The Five of Pentacles",
    "The Four of Pentacles",
    "The Three of Pentacles",
    "The Two of Pentacles",
    "The Ace of Pentacles",
];

const TAROT_SIDES: [(&str, &str); TAROT_MAX] = [
    /* Major Arcana */
    (
        "innocence, new beginnings, free spirit",
        "recklessness, taken advantage of, inconsideration",
    ),
    (
        "willpower, desire, creation, manifestation",
        "trickery, illusions, out of touch",
    ),
    (
        "intuitive, unconscious, inner voice",
        "lack of center, lost inner voice, repressed feelings",
    ),
    (
        "motherhood, fertility, nature",
        "dependence, smothering, emptiness, nosiness",
    ),
    (
        "authority, structure, control, fatherhood",
        "tyranny, rigidity, coldness",
    ),
    (
        "tradition, conformity, morality, ethics",
        "rebellion, subversiveness, new approaches",
    ),
    (
        "partnerships, duality, union",
        "loss of balance, one-sidedness, disharmony",
    ),
    (
        "direction, control, willpower",
        "lack of control, lack of direction, aggression",
    ),
    (
        "inner strength, bravery, compassion, focus",
        "self doubt, weakness, insecurity",
    ),
    (
        "contemplation, search for truth, inner guidance",
        "loneliness, isolation, lost your way",
    ),
    (
        "change, cycles, inevitable fate",
        "no control, clinging to control, bad luck",
    ),
    (
        "cause and effect, clarity, truth",
        "dishonesty, unaccountability, unfairness",
    ),
    (
        "sacrifice, release, martyrdom",
        "stalling, needless sacrifice, fear of sacrifice",
    ),
    (
        "end of cycle, beginnings, change, metamorphosis",
        "fear of change, holding on, stagnation, decay",
    ),
    (
        "middle path, patience, finding meaning",
        "extremes, excess, lack of balance",
    ),
    (
        "addiction, materialism, playfulness",
        "freedom, release, restoring control",
    ),
    (
        "sudden upheaval, broken pride, disaster",
        "disaster avoided, delayed disaster, fear of suffering",
    ),
    (
        "hope, faith, rejuvenation",
        "faithlessness, discouragement, insecurity",
    ),
    (
        "unconscious, illusions, intuition",
        "confusion, fear, misinterpretation",
    ),
    (
        "joy, success, celebration, positivity",
        "negativity, depression, sadness",
    ),
    (
        "reflection, reckoning, awakening",
        "lack of self awareness, doubt, self loathing",
    ),
    (
        "fulfillment, harmony, completion",
        "incompletion, no closure",
    ),
    /* Cups */
    (
        "compassion, control, balance",
        "coldness, moodiness, bad advice",
    ),
    (
        "compassion, calm, comfort",
        "martyrdom, insecurity, dependence",
    ),
    (
        "following the heart, idealist, romantic",
        "moodiness, disappointment",
    ),
    (
        "happy surprise, dreamer, sensitivity",
        "emotional immaturity, insecurity, disappointment",
    ),
    (
        "inner happiness, fulfillment, dreams coming true",
        "shattered dreams, broken family, domestic disharmony",
    ),
    (
        "satisfaction, emotional stability, luxury",
        "lack of inner joy, smugness, dissatisfaction",
    ),
    (
        "walking away, disillusionment, leaving behind",
        "avoidance, fear of change, fear of loss",
    ),
    (
        "searching for purpose, choices, daydreaming",
        "lack of purpose, diversion, confusion",
    ),
    (
        "familiarity, happy memories, healing",
        "moving forward, leaving home, independence",
    ),
    (
        "loss, grief, self-pity",
        "acceptance, moving on, finding peace",
    ),
    (
        "apathy, contemplation, disconnectedness",
        "sudden awareness, choosing happiness, acceptance",
    ),
    (
        "friendship, community, happiness",
        "overindulgence, gossip, isolation",
    ),
    (
        "unity, partnership, connection",
        "imbalance, broken communication, tension",
    ),
    (
        "new feelings, spirituality, intuition",
        "emotional loss, blocked creativity, emptiness",
    ),
    /* Swords */
    (
        "head over heart, discipline, truth",
        "manipulative, cruel, weakness",
    ),
    (
        "complexity, perceptiveness, clear mindedness",
        "cold hearted, cruel, bitterness",
    ),
    (
        "action, impulsiveness, defending beliefs",
        "no direction, disregard for consequences, unpredictability",
    ),
    (
        "curiosity, restlessness, mental energy",
        "deception, manipulation, all talk",
    ),
    (
        "failure, collapse, defeat",
        "can't get worse, only upwards, inevitable end",
    ),
    (
        "anxiety, hopelessness, trauma",
        "hope, reaching out, despair",
    ),
    (
        "imprisonment, entrapment, self-victimization",
        "self acceptance, new perspective, freedom",
    ),
    (
        "deception, trickery, tactics and strategy",
        "coming clean, rethinking approach, deception",
    ),
    (
        "transition, leaving behind, moving on",
        "emotional baggage, unresolved issues, resisting transition",
    ),
    (
        "unbridled ambition, win at all costs, sneakiness",
        "lingering resentment, desire to reconcile, forgiveness",
    ),
    (
        "rest, restoration, contemplation",
        "restlessness, burnout, stress",
    ),
    (
        "heartbreak, suffering, grief",
        "recovery, forgiveness, moving on",
    ),
    (
        "difficult choices, indecision, stalemate",
        "lesser of two evils, no right choice, confusion",
    ),
    (
        "breakthrough, clarity, sharp mind",
        "confusion, brutality, chaos",
    ),
    /* Wands */
    (
        "big picture, leader, overcoming challenges",
        "impulsive, overbearing, unachievable expectations",
    ),
    (
        "courage, determination, joy",
        "selfishness, jealousy, insecurities",
    ),
    (
        "action, adventure, fearlessness",
        "anger, impulsiveness, recklessness",
    ),
    (
        "exploration, excitement, freedom",
        "lack of direction, procrastination, creating conflict",
    ),
    (
        "accomplishment, responsibility, burden",
        "inability to delegate, overstressed, burnt out",
    ),
    (
        "resilience, grit, last stand",
        "exhaustion, fatigue, questioning motivations",
    ),
    (
        "rapid action, movement, quick decisions",
        "panic, waiting, slowdown",
    ),
    (
        "perseverance, defensive, maintaining control",
        "give up, destroyed confidence, overwhelmed",
    ),
    (
        "victory, success, public reward",
        "excess pride, lack of recognition, punishment",
    ),
    (
        "competition, rivalry, conflict",
        "avoiding conflict, respecting differences",
    ),
    (
        "community, home, celebration",
        "lack of support, transience, home conflicts",
    ),
    (
        "looking ahead, expansion, rapid growth",
        "obstacles, delays, frustration",
    ),
    (
        "planning, making decisions, leaving home",
        "fear of change, playing safe, bad planning",
    ),
    (
        "creation, willpower, inspiration, desire",
        "lack of energy, lack of passion, boredom",
    ),
    /* Pentacles */
    (
        "abundance, prosperity, security",
        "greed, indulgence, sensuality",
    ),
    (
        "practicality, creature comforts, financial security",
        "self-centeredness, jealousy, smothering",
    ),
    (
        "efficiency, hard work, responsibility",
        "laziness, obsessiveness, work without reward",
    ),
    (
        "ambition, desire, diligence",
        "lack of commitment, greediness, laziness",
    ),
    (
        "legacy, culmination, inheritance",
        "fleeting success, lack of stability, lack of resources",
    ),
    (
        "fruits of labor, rewards, luxury",
        "reckless spending, living beyond means, false success",
    ),
    (
        "apprenticeship, passion, high standards",
        "lack of passion, uninspired, no motivation",
    ),
    (
        "hard work, perseverance, diligence",
        "work without results, distractions, lack of rewards",
    ),
    (
        "charity, generosity, sharing",
        "strings attached, stinginess, power and domination",
    ),
    (
        "need, poverty, insecurity",
        "recovery, charity, improvement",
    ),
    (
        "conservation, frugality, security",
        "greediness, stinginess, possessiveness",
    ),
    (
        "teamwork, collaboration, building",
        "lack of teamwork, disorganized, group conflict",
    ),
    (
        "balancing decisions, priorities, adapting to change",
        "loss of balance, disorganized, overwhelmed",
    ),
    (
        "opportunity, prosperity, new venture",
        "lost opportunity, missed chance, bad investment",
    ),
];
