//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/main.rs: Main entry point
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! syd: seccomp and landlock based application sandbox with support for namespaces
//! Main entry point.

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    collections::HashSet,
    env,
    io::Write,
    os::unix::{ffi::OsStrExt, process::CommandExt},
    path::Path,
    process::{Command, ExitCode},
    str::FromStr,
};

use anyhow::{bail, Context, Error};
use getargs::{Opt, Options};
use libseccomp::{ScmpSyscall, ScmpVersion};
use nix::{
    errno::Errno,
    unistd::{getgid, getpid, getuid, Pid},
};
use once_cell::sync::Lazy;
use syd::{
    config::*,
    error,
    hook::{ExportMode, Supervisor},
    info,
    landlock::{
        path_beneath_rules, Access, AccessFs, RestrictionStatus, Ruleset, RulesetAttr,
        RulesetCreatedAttr, RulesetError, RulesetStatus, ABI,
    },
    sandbox,
    unshare::{GidMap, Namespace, UidMap},
    warn,
};

static SYD_VERSION: Lazy<&'static str> = Lazy::new(|| {
    if env!("SYD_GITHEAD").is_empty() {
        env!("CARGO_PKG_VERSION")
    } else {
        env!("SYD_GITHEAD")
    }
});

// This line will include the content of "esyd.sh" at compile-time
const ESYD: &str = include_str!("esyd.sh");

fn help() {
    let libapi = libseccomp::get_api();
    let libver = ScmpVersion::current().ok();
    let (libmaj, libmin, libmic) = libver
        .map(|v| (v.major, v.minor, v.micro))
        .unwrap_or((0, 0, 0));
    let native = syd::seccomp_arch_native_name().unwrap_or("?");

    println!(
        "SydB☮x {} libsecc☮mp v{}.{}.{} api:{} arch:{}
seccomp and landlock based application sandbox with support for namespaces
Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
SPDX-License-Identifier: GPL-3.0-or-later

# Usage

syd [-hvcl]
    [--lock] [--root]
    [-m command...] [-C path...] [-p name...]
    [--unshare-mount,uts,ipc,user,pid,net,cgroup]
    [-A arg0] [--map-root]
    [--] {{command [arg...]}}
syd --check
syd --export bpf|pfc
syd --list allow|deny|hook|ioctl|prctl
syd --print name
syd --arch
syd --errno number|name-regex
syd --read path
syd --syscall number|name-regex
syd --sh
syd --test [<name-regex>|<number>|<number>..<number>]..
syd exec {{command [arg...]}}
syd log

-h            -- Show usage and exit
-v            -- Show version and exit
-c            -- Causes command to be executed under a shell with the \"user\" profile
                 Login shell compatibility
-l            -- Ignored
                 Login shell compatibility
--lock        -- Lock magic commands
--root        -- Return 0 as user/group id, aka \"fakeroot\" mode
-m command    -- Run a magic command during init, may be repeated
-C path       -- Path to the configuration file to use, may be repeated
-p name       -- Name of the sandbox profile to use, may be repeated

--check       -- Exit with success if the process is running under SydB☮x
--export mode -- Export seccomp rules in given format to standard output and exit
--list set    -- Print the names of the system calls which belong to the given set and exit
                 If set is \"ioctl\", print the list of allowlisted ioctl requests
                 If set is \"prctl\", print the list of allowlisted prctl options
--print name  -- Print out the rules of the given sandbox profile and exit
--arch        -- Print the name of the libsecc☮mp native architecture and exit
--errno arg   -- Given a number, print the matching errno name and exit
                 Given a regex, print case-insensitively matching errno names and exit
--read path   -- Print the canonicalized path name followed by a newline and exit
--syscall arg -- Given a number, print the matching syscall name and exit
                 Given a regex, print case-insensitively matching syscall names and exit
--sh          -- Output a shell script which defines \"esyd\" the SydB☮x helper function
--test        -- Run integration tests and exit
                 Requires syd-test & syd-test-do programs to be in PATH.

-M, --unshare-mount   -- Unshare mount namespace
-S, --unshare-uts     -- Unshare UTS namespace
-I, --unshare-ipc     -- Unshare IPC namespace
-U, --unshare-user    -- Unshare user namespace
-P, --unshare-pid     -- Unshare pid namespace
-N, --unshare-net     -- Unshare net namespace
-G, --unshare-cgroup  -- Unshare cgroup namespace

-A, --arg0 alias      -- Set alias of the command
                         Passed as `argv[0]` to the program
-R, --map-root        -- Map current user to root in the sandbox
                         Implies --unshare-user

* syd exec may be used to construct a sandbox command to execute a
process outside the sandbox. See the description of cmd/exec
command in the README for more information.
* syd log may be used to access sandbox logs using journalctl.

# Profiles

1. paludis: Used by the Paludis package mangler.
2. noipv4: Disables IPv4 connectivity.
3. noipv6: Disables Ipv6 connectivity.
4. silent: Silences all access violations.
5. user: Allows user-specific directories and connections,
         and parses the following files if they exist:
         1. /etc/user.syd-3
         2. ~/.user.syd-3

When invoked without arguments, the current shell is executed
under sandbox with the user profile.

# Environment

SYD_LOG: Set log level. See the Enabling Logging¹ section of
 env-logger crate documentation for more information.
SYD_NPROC: Set the number of system call handler threads.
           By default this is equal to the the number of CPUs on the machine.
SYD_NO_SYSLOG: Disable logging to syslog(3). By default logs of
               severity Warn and higher are logged to syslog(3).
SYD_NO_CROSS_MEMORY_ATTACH: Disable cross memory attach,
                            and fallback to /proc/$pid/mem.
SYD_FAKEROOT: Equivalent to passing --root
SYD_SH: Pick the shell to spawn when invoked as a login shell
        Defaults to /bin/bash
SYD_UNSHARE_MOUNT: Equivalent to passing --unshare-mount
SYD_UNSHARE_UTS: Equivalent to passing --unshare-uts
SYD_UNSHARE_IPC: Equivalent to passing --unshare-ipc
SYD_UNSHARE_PID: Equivalent to passing --unshare-pid
SYD_UNSHARE_NET: Equivalent to passing --unshare-net
SYD_UNSHARE_CGROUP: Equivalent to passing --unshare-cgroup

# Exit Codes

Sydb☮x exits with the same exit code as the sandbox process
itself. If the sandbox process exits with a signal, Sydb☮x exits
with 14 which stands for EFAULT. In case there was an error in
spawning or waiting for the sandbox process, Sydb☮x exits with
errno indicating the error condition. E.g. syd true\" returns 0,
\"syd false\" return 1, and \"syd -- syd true\" returns 16 which
stands for EBUSY which stands for \"Device or resource busy\"
indicating there is already a secure computing filter
loaded. tl;dr Sydb☮x won't run under Sydb☮x, similarly many
process inspection tools such as \"ltrace\", \"strace\", or
\"gdb\" won't work under Sydb☮x. Thus the sandbox process can
either be traced by attaching from outside the sandbox or
running the tracer in follow fork mode, e.g.
\"strace -f syd true\".

¹: https://docs.rs/env_logger/latest/env_logger/#enabling-logging",
        *SYD_VERSION, libmaj, libmin, libmic, libapi, native
    );
}

#[allow(clippy::cognitive_complexity)]
fn main() -> anyhow::Result<ExitCode> {
    // Step 1: Initialize environment.
    Supervisor::init_env()?;

    // Step 2: Parse CLI arguments.
    let mut argv_iter = argv::iter();
    #[allow(clippy::disallowed_methods)]
    let is_login = *argv_iter.next().unwrap().as_bytes().first().unwrap() == b'-';
    #[allow(clippy::disallowed_methods)]
    let mut opts =
        Options::new(argv_iter.map(|os| os.to_str().expect("Argument is not valid UTF-8!")));

    // Initialize options.
    let sh = env::var(ENV_SH).unwrap_or(SYD_SH.to_string());
    let mut arg0: Option<String> = None;
    let mut argv: Vec<String> = Vec::new();
    let mut test = false;
    let mut root = env::var(ENV_FAKEROOT).is_ok();
    let mut map_root = false;
    let mut export_mode: Option<ExportMode> = None;
    let mut list_set: Option<String> = None;
    let mut print_profile: Option<String> = None;
    let mut sandbox = sandbox::Sandbox::new();

    // Initialize namespace options from the environment.
    sandbox.set_unshare_mount(env::var(ENV_UNSHARE_MOUNT).is_ok());
    sandbox.set_unshare_uts(env::var(ENV_UNSHARE_UTS).is_ok());
    sandbox.set_unshare_ipc(env::var(ENV_UNSHARE_IPC).is_ok());
    sandbox.set_unshare_user(env::var(ENV_UNSHARE_USER).is_ok());
    sandbox.set_unshare_pid(env::var(ENV_UNSHARE_PID).is_ok());
    sandbox.set_unshare_net(env::var(ENV_UNSHARE_NET).is_ok());
    sandbox.set_unshare_cgroup(env::var(ENV_UNSHARE_CGROUP).is_ok());

    while let Some(opt) = opts.next_opt().context("calling Options::next")? {
        match opt {
            /*
             * Basic options
             */
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::from(0));
            }
            Opt::Short('v') | Opt::Long("version") => {
                let libapi = libseccomp::get_api();
                let libver =
                    ScmpVersion::current().context("Failed to determine libsecc☮mp version.")?;
                println!(
                    "SydB☮x {} libsecc☮mp v{}.{}.{} api:{} arch:{}",
                    *SYD_VERSION,
                    libver.major,
                    libver.minor,
                    libver.micro,
                    libapi,
                    syd::seccomp_arch_native_name().unwrap_or("?"),
                );
                return Ok(ExitCode::from(0));
            }

            /*
             * Sandbox options
             */
            Opt::Long("check") => {
                return Ok(ExitCode::from(if syd::syd_enabled() { 0 } else { 1 }));
            }
            Opt::Long("sh") => {
                println!("{ESYD}");
                return Ok(ExitCode::from(0));
            }
            Opt::Long("export") => {
                export_mode = Some(ExportMode::from_str(
                    opts.value().context("--export requires an argument!")?,
                )?);
            }
            Opt::Long("list") => {
                list_set = Some(
                    opts.value()
                        .context("--list requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("print") => {
                print_profile = Some(
                    opts.value()
                        .context("--print requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("arch") => {
                if let Some(name) = syd::seccomp_arch_native_name() {
                    println!("{name}");
                    return Ok(ExitCode::from(0));
                } else {
                    return Ok(ExitCode::from(1));
                }
            }
            Opt::Long("errno") => {
                let val = opts.value().context("--errno requires an argument!")?;
                match val.parse::<u16>() {
                    Ok(0) => {
                        return Ok(ExitCode::from(1));
                    }
                    Ok(num) => {
                        // number -> name
                        let errno = Errno::from_i32(i32::from(num));
                        if errno == Errno::UnknownErrno {
                            return Ok(ExitCode::from(1));
                        }
                        let estr = errno.to_string();
                        let mut iter = estr.split(": ");
                        let name = iter.next().unwrap_or("?");
                        let desc = iter.next().unwrap_or("?");
                        println!("{num}\t{name}\t{desc}");
                    }
                    Err(_) => {
                        match regex::RegexBuilder::new(val)
                            .unicode(false)
                            .case_insensitive(true)
                            .build()
                        {
                            Ok(pattern) => {
                                // regex -> [number]
                                let mut ok = false;
                                for errno in (1..u8::MAX).map(|n| Errno::from_i32(i32::from(n))) {
                                    if errno == Errno::UnknownErrno {
                                        continue;
                                    }
                                    let estr = errno.to_string();
                                    let mut iter = estr.split(": ");
                                    let name = iter.next().unwrap_or("?");
                                    let desc = iter.next().unwrap_or("?");
                                    if pattern.is_match(name) {
                                        println!("{}\t{}\t{}", errno as i32, name, desc);
                                        ok = true;
                                    }
                                }
                                if !ok {
                                    return Ok(ExitCode::from(1));
                                }
                            }
                            Err(error) => {
                                eprintln!("Invalid errno regex \"{val}\": {error}");
                                return Ok(ExitCode::from(1));
                            }
                        }
                    }
                }
                return Ok(ExitCode::from(0));
            }
            Opt::Long("read") => {
                let path = Path::new(opts.value().context("--read requires an argument!")?);
                let path = if path.is_relative() {
                    let cdir = syd::fs::canonicalize(
                        "/proc/self/cwd",
                        true,
                        syd::fs::MissingHandling::Missing,
                    )
                    .context("Failed to canonicalize /proc/self/cwd")?;
                    cdir.join(path)
                } else {
                    path.to_path_buf()
                };
                let path = syd::fs::canonicalize(path, true, syd::fs::MissingHandling::Missing)
                    .context("Failed to canonicalize path!")?;
                let path = path.as_os_str().as_bytes();
                std::io::stdout()
                    .write_all(path)
                    .context("Error writing canonicalized path to standard output!")?;
                println!();
                return Ok(ExitCode::from(0));
            }
            Opt::Long("syscall") => {
                let val = opts.value().context("--syscall requires an argument!")?;
                match val.parse::<i32>() {
                    Ok(num) => {
                        let syscall = ScmpSyscall::from(num);
                        if let Ok(name) = syscall.get_name() {
                            println!("{num}\t{name}");
                        } else {
                            return Ok(ExitCode::from(1));
                        }
                    }
                    Err(_) => {
                        match regex::RegexBuilder::new(val)
                            .unicode(false)
                            .case_insensitive(true)
                            .build()
                        {
                            Ok(pattern) => {
                                let mut ok = false;
                                for (num, name) in (0..4096)
                                    .map(|n| {
                                        (n, ScmpSyscall::from(n).get_name().unwrap_or_default())
                                    })
                                    .filter(|(_, name)| !name.is_empty())
                                {
                                    if pattern.is_match(&name) {
                                        println!("{num}\t{name}");
                                        ok = true;
                                    }
                                }
                                if !ok {
                                    return Ok(ExitCode::from(1));
                                }
                            }
                            Err(error) => {
                                eprintln!("Invalid syscall regex \"{val}\": {error}");
                                return Ok(ExitCode::from(1));
                            }
                        }
                    }
                }
                return Ok(ExitCode::from(0));
            }
            Opt::Long("test") => {
                test = true;
            }
            Opt::Long("lock") => {
                sandbox.lock();
            }
            Opt::Long("root") => {
                root = true;
            }
            Opt::Short('m') | Opt::Long("magic") => {
                let cmd = opts.value().context("--magic requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "cfg": cmd, "err": "LOCK");
                } else {
                    sandbox
                        .config(cmd)
                        .context(format!("Failed to execute magic command `{cmd}'."))?;
                }
            }
            Opt::Short('l') | Opt::Long("login") => {
                // Ignored, login shell compatibility
            }
            Opt::Short('c') => {
                // syd -c command args... akin to bash -c command args..
                if argv.is_empty() {
                    sandbox
                        .parse_profile("user")
                        .context("Unable to parse user profile!")?;
                    argv.push(sh.clone());
                    argv.push("-c".to_string());
                }
            }
            Opt::Short('C') | Opt::Long("config") => {
                let path = opts.value().context("--config requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "path": path, "err": "LOCK");
                } else {
                    sandbox
                        .parse_config_file(path)
                        .context(format!("Failed to parse configuration file `{path}'."))?;
                }
            }
            Opt::Short('p') | Opt::Long("profile") => {
                let profile = opts.value().context("--profile requires an argument!")?;
                if sandbox.locked() {
                    warn!("ctx": "config", "pid": getpid().as_raw(), "prof": profile, "err": "LOCK");
                } else {
                    sandbox.parse_profile(profile).context(format!(
                        "Failed to parse configuration profile `{profile}'."
                    ))?;
                }
            }

            /*
             * Unshare options
             */
            Opt::Short('M') | Opt::Long("unshare-mount") => {
                sandbox.set_unshare_mount(true);
            }
            Opt::Short('S') | Opt::Long("unshare-uts") => {
                sandbox.set_unshare_uts(true);
            }
            Opt::Short('I') | Opt::Long("unshare-ipc") => {
                sandbox.set_unshare_ipc(true);
            }
            Opt::Short('U') | Opt::Long("unshare-user") => {
                sandbox.set_unshare_user(true);
            }
            Opt::Short('P') | Opt::Long("unshare-pid") => {
                sandbox.set_unshare_pid(true);
            }
            Opt::Short('N') | Opt::Long("unshare-net") => {
                sandbox.set_unshare_net(true);
            }
            Opt::Short('G') | Opt::Long("unshare-cgroup") => {
                sandbox.set_unshare_cgroup(true);
            }
            Opt::Short('A') | Opt::Long("arg0") => {
                let val = opts
                    .value()
                    .context("--arg0 requires an argument!")?
                    .to_string();
                arg0 = Some(val);
            }
            Opt::Short('R') | Opt::Long("map-root") => {
                map_root = true;
                sandbox.set_unshare_user(true);
            }
            Opt::Short(c) => {
                bail!("Invalid option `-{c}'!");
            }
            Opt::Long(c) => {
                bail!("Invalid option `--{c}'!");
            }
        }
    }

    if test {
        Command::new("syd-test").args(opts.positionals()).exec();
    } else if root {
        sandbox.set_root(true);
    }

    // List set or Print profile if requested, and exit.
    match (list_set.as_deref(), print_profile.as_deref()) {
        (Some(_), Some(_)) => {
            eprintln!("The options --list and --print are mutually exclusive!");
            return Ok(ExitCode::from(1));
        }
        (Some("ioctl"), _) => {
            for (_, name) in &ALLOWLIST_IOCTL {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("prctl"), _) => {
            for (_, name) in &ALLOWLIST_PRCTL {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("allow"), _) => {
            for name in &SAFE_SYSCALLS {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("hook"), _) => {
            for name in &HOOK_SYSCALLS {
                println!("{name}");
            }
            return Ok(ExitCode::from(0));
        }
        (Some("deny"), _) => {
            let mut syscall_set: HashSet<_> = SAFE_SYSCALLS.into_iter().map(String::from).collect();
            for syscall in &HOOK_SYSCALLS {
                syscall_set.insert(syscall.to_string());
            }
            for syscall_number in 0..=600 {
                let syscall = ScmpSyscall::from(syscall_number);
                if let Ok(name) = syscall.get_name() {
                    if !syscall_set.contains(&name) {
                        println!("{name}");
                    }
                }
            }
            return Ok(ExitCode::from(0));
        }
        (Some(set), _) => {
            eprintln!("No such set: '{set}'");
            return Ok(ExitCode::from(1));
        }
        (_, Some("paludis")) => {
            println!("# SydB☮x profile: Paludis");
            println!("# Number of rules: {PROFILE_PALUDIS_SIZE}");
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_PALUDIS.iter().take(PROFILE_PALUDIS_SIZE) {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("noipv4")) => {
            println!("# SydB☮x profile: NoIpv4");
            println!("# Number of rules: {PROFILE_NOIPV4_SIZE}");
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV4.iter().take(PROFILE_NOIPV4_SIZE) {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("noipv6")) => {
            println!("# SydB☮x profile: NoIpv6");
            println!("# Number of rules: {PROFILE_NOIPV6_SIZE}");
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_NOIPV6.iter().take(PROFILE_NOIPV6_SIZE) {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("silent")) => {
            println!("# SydB☮x profile: Silent");
            println!("# Number of rules: {PROFILE_SILENT_SIZE}");
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in PROFILE_SILENT.iter().take(PROFILE_SILENT_SIZE) {
                println!("{command}");
            }
            return Ok(ExitCode::from(0));
        }
        (_, Some("user")) => {
            let uid = nix::unistd::getuid();
            let name = syd::get_user_name(uid);
            let mut home = syd::get_user_home(&name);

            println!("# SydB☮x profile: User \"{name}\"");
            println!("# Number of rules: {}", PROFILE_USER_SIZE + 9);
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");

            // Step 1: Static configuration defined at compile time.
            for command in PROFILE_USER.iter().take(PROFILE_USER_SIZE) {
                println!("{command}");
            }

            // Step 2: Dynamic, user-specific configuration.
            // Keep in sync with Sandbox::parse_profile()
            if !home.ends_with('/') {
                home.push('/');
            }

            println!("allowlist/lock/write+{home}");
            println!("allowlist/read+{home}***");
            println!("allowlist/stat+{home}***");
            println!("allowlist/write+{home}**");
            println!("allowlist/exec+{home}**");
            println!("allowlist/net/bind+{home}**");
            println!("allowlist/net/connect+{home}**");
            println!("allowlist/read+/run/user/{uid}/**");
            println!("allowlist/write+/run/user/{uid}/**");

            return Ok(ExitCode::from(0));
        }
        (_, Some(name)) => {
            eprintln!("No such profile: '{name}'");
            return Ok(ExitCode::from(1));
        }
        (None, None) => {
            // Do nothing, fall through.
        }
    };

    // Step 3: Prepare the command to execute, which may be a login shell.
    if argv.is_empty() {
        argv.extend(opts.positionals().map(|s| s.to_string()));
    } else {
        // -c command args...
        argv.push(opts.positionals().collect::<Vec<_>>().join(" "));
    };
    let mut argv = argv.into_iter();
    let argv0 = match (export_mode, argv.next(), is_login) {
        (Some(_), _, false) => "true".to_string(),
        (_, Some(ref cmd), false) if cmd == "log" => {
            Command::new("journalctl")
                .args(["SYSLOG_IDENTIFIER=syd"])
                .exec();
            return Ok(ExitCode::from(1));
        }
        (_, Some(ref cmd), false) if cmd == "exec" => {
            // Split the arguments using the ASCII Unit Separator character
            let args = argv.collect::<Vec<_>>().join("\x1F");

            // Format it using /dev/syd/cmd/exec!<concatenated-path>
            let path = format!("/dev/syd/cmd/exec!{args}");
            print!("{path}");
            return Ok(ExitCode::from(0));
        }
        (_, Some(argv0), false) => argv0,
        (_, None, false) | (_, _, true) => {
            sandbox
                .parse_profile("user")
                .context("Unable to parse user profile!")?;
            arg0 = Some(
                Path::new(&sh)
                    .file_name()
                    .map(|n| format!("-{}", n.to_string_lossy()))
                    .unwrap_or("-".to_string()),
            );
            sh.to_string()
        }
    };

    // Step 4: Set up the Landlock sandbox if requested. Note,
    // we set it up here before spawning the child so as to
    // include the SydB☮x sandbox threads into the sandbox as
    // well. This is done for added security.
    // Note, Landlock errors are not fatal.
    if let Some((mut path_ro, mut path_rw)) = sandbox.collect_landlock() {
        path_ro.sort();
        path_rw.sort();
        let abi = ABI::V3;

        // A helper function to wrap the operations and reduce duplication
        fn landlock_operation(
            path_ro: &[String],
            path_rw: &[String],
            abi: ABI,
        ) -> Result<RestrictionStatus, RulesetError> {
            let ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
            let created_ruleset = ruleset.create()?;
            let ro_rules = path_beneath_rules(path_ro, AccessFs::from_read(abi));
            let updated_ruleset = created_ruleset.add_rules(ro_rules)?;
            let rw_rules = path_beneath_rules(path_rw, AccessFs::from_all(abi));
            let final_ruleset = updated_ruleset.add_rules(rw_rules)?;
            final_ruleset.restrict_self().map_err(RulesetError::from)
        }

        match landlock_operation(&path_ro, &path_rw, abi) {
            Ok(status) => {
                info!("ctx": "landlock",
                "pid": Pid::this().as_raw(),
                "path_ro": path_ro,
                "path_rw": path_rw,
                "status": match status.ruleset {
                    // The FullyEnforced case must be tested by the developer.
                    RulesetStatus::FullyEnforced => "fully_enforced",
                    RulesetStatus::PartiallyEnforced => "partially_enforced",
                    // Users should be warned that they are not protected.
                    RulesetStatus::NotEnforced => "not_enforced",
                });
            }
            Err(error) => {
                error!("ctx": "landlock", "err": error.to_string());
            }
        }
    }

    // Step 5: Initialize sandbox supervisor.
    let nproc = if let Ok(var) = env::var(ENV_NPROC) {
        var.parse::<usize>()
            .context("Failed to parse number of threads from the SYD_NPROC environment variable.")?
    } else {
        num_cpus::get()
    };
    sandbox
        .build_globsets()
        .context("Failed to build GlobSets.")?;
    let namespaces = sandbox.namespaces();
    let mut supervisor = Supervisor::new(sandbox, nproc, export_mode).context(format!(
        "Error creating sandbox with {} threads and export mode {:?}.",
        nproc, export_mode
    ))?;
    supervisor.init();

    let mut command = syd::unshare::Command::new(argv0);
    command
        .args(argv)
        .stderr(syd::unshare::Stdio::inherit())
        .stdin(syd::unshare::Stdio::inherit())
        .stdout(syd::unshare::Stdio::inherit())
        .unshare(&namespaces);
    if let Some(arg0) = arg0 {
        command.arg0(arg0);
    }
    if namespaces.contains(&Namespace::User) {
        // set up id mappings
        let uid = getuid().as_raw();
        let gid = getgid().as_raw();

        // create the UID and GID mappings.
        let uid_maps = vec![
            UidMap {
                inside_uid: if map_root { 0 } else { uid },
                outside_uid: uid,
                count: 1,
            }, // Map the current user.
        ];
        let gid_maps = vec![
            GidMap {
                inside_gid: if map_root { 0 } else { gid },
                outside_gid: gid,
                count: 1,
            }, // Map the current group.
        ];

        // Set the UID and GID mappings for the command.
        command.set_id_maps(uid_maps, gid_maps);
    }

    // Step 6: Spawn the program under sandbox.
    let (pid, thread_handle, pool) = match supervisor.spawn(&mut command) {
        Ok((pid, thread_handle, pool)) => (pid, thread_handle, pool),
        Err(error) => {
            let errno = Errno::last() as u8;
            let error = error.context("Failed to spawn command under sandbox.");
            eprintln!("{error:?}");
            return Ok(ExitCode::from(errno));
        }
    };

    // Step 7: Wait for the process to exit and return the same error code.
    let code: u8 = match Supervisor::wait(pid, thread_handle, pool) {
        Ok(status) => status.code().unwrap_or(nix::libc::EFAULT),
        Err(errno) => {
            let error = Error::new(errno).context("Failed to wait for sandboxed process.");
            eprintln!("{error:?}");
            errno as i32
        }
    }
    .try_into()
    .unwrap_or(127);
    Ok(ExitCode::from(code))
}
