//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/test/test.rs: Integration tests
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(clippy::disallowed_methods)]

use std::env;

use nix::sys::stat::{umask, Mode};

use crate::{assert, assert_eq, assert_ne, ignore, util::*, KERNEL_VERSION};

/// Represents a test case.
pub type Test<'a> = (&'a str, fn() -> TestResult);

macro_rules! test_entry {
    ($func:expr) => {
        (stringify!($func), $func)
    };
}

/// Number of integration tests
pub const TEST_COUNT: usize = 169;
/// List of integration tests.
pub const TESTS: [Test; TEST_COUNT] = [
    test_entry!(test_syd_true_returns_success),
    test_entry!(test_syd_true_returns_success_with_many_processes),
    test_entry!(test_syd_true_returns_success_with_many_threads),
    test_entry!(test_syd_false_returns_failure),
    test_entry!(test_syd_true_returns_failure_with_many_processes),
    test_entry!(test_syd_true_returns_failure_with_many_threads),
    test_entry!(test_syd_whoami_returns_root),
    test_entry!(test_syd_lock),
    test_entry!(test_syd_lock_exec),
    test_entry!(test_syd_check),
    test_entry!(test_syd_read_sandbox_open_allow),
    test_entry!(test_syd_read_sandbox_open_deny),
    test_entry!(test_syd_stat_sandbox_chdir_allow),
    test_entry!(test_syd_stat_sandbox_chdir_hide),
    test_entry!(test_syd_stat_sandbox_stat_allow),
    test_entry!(test_syd_stat_sandbox_stat_hide),
    test_entry!(test_syd_stat_sandbox_getdents_allow),
    test_entry!(test_syd_stat_sandbox_getdents_hide),
    test_entry!(test_syd_write_sandbox_open_allow),
    test_entry!(test_syd_write_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_open_allow),
    test_entry!(test_syd_exec_sandbox_open_deny),
    test_entry!(test_syd_network_sandbox_connect_ipv4_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv4_deny),
    test_entry!(test_syd_network_sandbox_connect_ipv6_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv6_deny),
    test_entry!(test_syd_network_sandbox_allow_successful_bind_ipv4_failure),
    test_entry!(test_syd_network_sandbox_allow_successful_bind_ipv4_success),
    test_entry!(test_syd_network_sandbox_allow_successful_bind_ipv6_failure),
    test_entry!(test_syd_network_sandbox_allow_successful_bind_ipv6_success),
    test_entry!(test_syd_symlink_toctou),
    test_entry!(test_syd_symlinkat_toctou),
    test_entry!(test_syd_ptrmod_toctou_open),
    test_entry!(test_syd_ptrmod_toctou_creat),
    test_entry!(test_syd_io_uring_escape),
    test_entry!(test_syd_opath_escape),
    test_entry!(test_syd_devfd_escape_chdir),
    test_entry!(test_syd_devfd_escape_chdir_relpath_1),
    test_entry!(test_syd_devfd_escape_chdir_relpath_2),
    test_entry!(test_syd_devfd_escape_chdir_relpath_3),
    test_entry!(test_syd_devfd_escape_chdir_relpath_4),
    test_entry!(test_syd_devfd_escape_chdir_relpath_5),
    test_entry!(test_syd_devfd_escape_chdir_relpath_6),
    test_entry!(test_syd_devfd_escape_chdir_relpath_7),
    test_entry!(test_syd_devfd_escape_chdir_relpath_8),
    test_entry!(test_syd_devfd_escape_chdir_relpath_9),
    test_entry!(test_syd_devfd_escape_chdir_relpath_10),
    test_entry!(test_syd_devfd_escape_chdir_relpath_11),
    test_entry!(test_syd_devfd_escape_chdir_relpath_12),
    test_entry!(test_syd_devfd_escape_chdir_relpath_13),
    test_entry!(test_syd_devfd_escape_chdir_relpath_14),
    test_entry!(test_syd_devfd_escape_chdir_relpath_15),
    test_entry!(test_syd_devfd_escape_chdir_relpath_16),
    test_entry!(test_syd_devfd_escape_chdir_relpath_17),
    test_entry!(test_syd_devfd_escape_chdir_relpath_18),
    test_entry!(test_syd_devfd_escape_chdir_relpath_19),
    test_entry!(test_syd_devfd_escape_chdir_relpath_20),
    test_entry!(test_syd_devfd_escape_open),
    test_entry!(test_syd_devfd_escape_open_relpath_1),
    test_entry!(test_syd_devfd_escape_open_relpath_2),
    test_entry!(test_syd_devfd_escape_open_relpath_3),
    test_entry!(test_syd_devfd_escape_open_relpath_4),
    test_entry!(test_syd_devfd_escape_open_relpath_5),
    test_entry!(test_syd_devfd_escape_open_relpath_6),
    test_entry!(test_syd_devfd_escape_open_relpath_7),
    test_entry!(test_syd_devfd_escape_open_relpath_8),
    test_entry!(test_syd_devfd_escape_open_relpath_9),
    test_entry!(test_syd_devfd_escape_open_relpath_10),
    test_entry!(test_syd_devfd_escape_open_relpath_11),
    test_entry!(test_syd_devfd_escape_open_relpath_12),
    test_entry!(test_syd_devfd_escape_open_relpath_13),
    test_entry!(test_syd_devfd_escape_open_relpath_14),
    test_entry!(test_syd_devfd_escape_open_relpath_15),
    test_entry!(test_syd_devfd_escape_open_relpath_16),
    test_entry!(test_syd_devfd_escape_open_relpath_17),
    test_entry!(test_syd_devfd_escape_open_relpath_18),
    test_entry!(test_syd_devfd_escape_open_relpath_19),
    test_entry!(test_syd_devfd_escape_open_relpath_20),
    test_entry!(test_syd_procself_escape_chdir),
    test_entry!(test_syd_procself_escape_chdir_relpath_1),
    test_entry!(test_syd_procself_escape_chdir_relpath_2),
    test_entry!(test_syd_procself_escape_chdir_relpath_3),
    test_entry!(test_syd_procself_escape_chdir_relpath_4),
    test_entry!(test_syd_procself_escape_chdir_relpath_5),
    test_entry!(test_syd_procself_escape_chdir_relpath_6),
    test_entry!(test_syd_procself_escape_chdir_relpath_7),
    test_entry!(test_syd_procself_escape_chdir_relpath_8),
    test_entry!(test_syd_procself_escape_chdir_relpath_9),
    test_entry!(test_syd_procself_escape_chdir_relpath_10),
    test_entry!(test_syd_procself_escape_chdir_relpath_11),
    test_entry!(test_syd_procself_escape_chdir_relpath_12),
    test_entry!(test_syd_procself_escape_chdir_relpath_13),
    test_entry!(test_syd_procself_escape_chdir_relpath_14),
    test_entry!(test_syd_procself_escape_chdir_relpath_15),
    test_entry!(test_syd_procself_escape_chdir_relpath_16),
    test_entry!(test_syd_procself_escape_chdir_relpath_17),
    test_entry!(test_syd_procself_escape_chdir_relpath_18),
    test_entry!(test_syd_procself_escape_chdir_relpath_19),
    test_entry!(test_syd_procself_escape_chdir_relpath_20),
    test_entry!(test_syd_procself_escape_open),
    test_entry!(test_syd_procself_escape_open_relpath_1),
    test_entry!(test_syd_procself_escape_open_relpath_2),
    test_entry!(test_syd_procself_escape_open_relpath_3),
    test_entry!(test_syd_procself_escape_open_relpath_4),
    test_entry!(test_syd_procself_escape_open_relpath_5),
    test_entry!(test_syd_procself_escape_open_relpath_6),
    test_entry!(test_syd_procself_escape_open_relpath_7),
    test_entry!(test_syd_procself_escape_open_relpath_8),
    test_entry!(test_syd_procself_escape_open_relpath_9),
    test_entry!(test_syd_procself_escape_open_relpath_10),
    test_entry!(test_syd_procself_escape_open_relpath_11),
    test_entry!(test_syd_procself_escape_open_relpath_12),
    test_entry!(test_syd_procself_escape_open_relpath_13),
    test_entry!(test_syd_procself_escape_open_relpath_14),
    test_entry!(test_syd_procself_escape_open_relpath_15),
    test_entry!(test_syd_procself_escape_open_relpath_16),
    test_entry!(test_syd_procself_escape_open_relpath_17),
    test_entry!(test_syd_procself_escape_open_relpath_18),
    test_entry!(test_syd_procself_escape_open_relpath_19),
    test_entry!(test_syd_procself_escape_open_relpath_20),
    test_entry!(test_syd_procself_escape_relpath),
    test_entry!(test_syd_umask_bypass_077),
    test_entry!(test_syd_umask_bypass_277),
    test_entry!(test_syd_emulate_opath),
    test_entry!(test_syd_emulate_otmpfile),
    test_entry!(test_syd_honor_umask),
    test_entry!(test_syd_open_utf8_invalid),
    test_entry!(test_syd_exec_in_inaccessible_directory),
    test_entry!(test_syd_disallow_setuid),
    test_entry!(test_syd_disallow_setgid),
    test_entry!(test_syd_fstat_on_pipe),
    test_entry!(test_syd_fstat_on_socket),
    test_entry!(test_syd_fstat_on_deleted_file),
    test_entry!(test_syd_fstat_on_temp_file),
    test_entry!(test_syd_block_ioctl_tiocsti),
    test_entry!(test_syd_block_prctl_ptrace),
    test_entry!(test_syd_block_dev_random),
    test_entry!(test_syd_kill_during_syscall),
    test_entry!(test_syd_open_toolong_path),
    test_entry!(test_syd_open_null_path),
    test_entry!(test_syd_utimensat_null),
    test_entry!(test_syd_normalize_path),
    test_entry!(test_syd_path_resolution),
    test_entry!(test_syd_emulate_open_fifo),
    test_entry!(test_syd_remove_empty_path),
    test_entry!(test_syd_symlink_readonly_path),
    test_entry!(test_syd_open_trailing_slash),
    test_entry!(test_syd_openat_trailing_slash),
    test_entry!(test_syd_lstat_trailing_slash),
    test_entry!(test_syd_fstatat_trailing_slash),
    test_entry!(test_syd_mkdir_trailing_dot),
    test_entry!(test_syd_mkdirat_trailing_dot),
    test_entry!(test_syd_rmdir_trailing_slashdot),
    test_entry!(test_syd_fopen_supports_mode_e),
    test_entry!(test_syd_fopen_supports_mode_x),
    test_entry!(test_syd_link_no_symlink_deref),
    test_entry!(test_syd_link_posix),
    test_entry!(test_syd_linkat_posix),
    test_entry!(test_syd_getcwd_long),
    test_entry!(test_syd_creat_thru_dangling),
    test_entry!(test_syd_mkdirat_non_dir_fd),
    test_entry!(test_syd_blocking_udp4),
    test_entry!(test_syd_blocking_udp6),
    test_entry!(test_syd_close_on_exec),
    test_entry!(test_syd_open_exclusive_restart),
    test_entry!(test_syd_open_exclusive_repeat),
    test_entry!(test_syd_pty_io),
    test_entry!(test_syd_diff_dev_fd),
    test_entry!(test_syd_syscall_fuzz),
];

// Tests if `true` returns success under sandbox.
fn test_syd_true_returns_success() -> TestResult {
    let status = syd().arg("true").status().expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd().args(["--", "true"]).status().expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `syd` returns success for a sandbox running many processes,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_processes() -> TestResult {
    env::set_var("SYD_DO", "fork");
    let status = syd()
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `syd` returns success for a sandbox running many threads,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_threads() -> TestResult {
    env::set_var("SYD_DO", "thread");
    let status = syd()
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `false` returns failure under sandbox.
fn test_syd_false_returns_failure() -> TestResult {
    let status = syd().arg("false").status().expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0, "status:{status:?}");

    let status = syd().args(["--", "false"]).status().expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0, "status:{status:?}");

    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many processes,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_processes() -> TestResult {
    env::set_var("SYD_DO", "fork");
    let status = syd()
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 7);

    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many threads,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_threads() -> TestResult {
    env::set_var("SYD_DO", "thread");
    let status = syd()
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 7);

    Ok(())
}

// Tests if `whoami` returns `root` with `syd --root`.
fn test_syd_whoami_returns_root() -> TestResult {
    let output = syd()
        .args(["--root", "--", "whoami"])
        .output()
        .expect("execute syd");
    assert!(output.stdout.starts_with(b"root"));

    Ok(())
}

// Tests if `syd --lock` disables access to `/dev/syd`.
fn test_syd_lock() -> TestResult {
    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .args(["--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .args(["--lock", "--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if `lock:exec` locks the sandbox for all except the exec child.
fn test_syd_lock_exec() -> TestResult {
    eprintln!(r#"+ bash -c "test -e /dev/syd""#);
    let status = syd()
        .args(["-mlock:exec", "--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    eprintln!(r#"+ bash -c "( test -e /dev/syd )""#);
    let status = syd()
        .args([
            "-mlock:exec",
            "--",
            "/bin/bash",
            "-c",
            "( test -e /dev/syd )",
        ])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if syd --check works to check if process is running under SydB☮x.
fn test_syd_check() -> TestResult {
    eprintln!("+ syd --check");
    let status = syd().arg("--check").status().expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 1);

    eprintln!("+ syd -- syd --check");
    let status = syd()
        .args(["--", &SYD, "--check"])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if read sandboxing for open works to allow.
fn test_syd_read_sandbox_open_allow() -> TestResult {
    eprintln!("+ dd if=/dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "-mallowlist/read+/***",
            "-mdenylist/read+/dev/***",
            "-mallowlist/read+/dev/null",
            "--",
            "dd",
            "if=/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if read sandboxing for open works to deny.
fn test_syd_read_sandbox_open_deny() -> TestResult {
    eprintln!("+ cat /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "-mdenylist/read+/dev/null",
            "--",
            "cat",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if stat sandboxing for chdir works to allow.
fn test_syd_stat_sandbox_chdir_allow() -> TestResult {
    eprintln!("+ bash -c \"cd /dev\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdenylist/stat+/dev",
            "-mallowlist/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_chdir_hide() -> TestResult {
    eprintln!("+ cd /dev");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "-mdenylist/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev || exit 42",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 42);

    Ok(())
}

// Tests if stat sandboxing for stat works to allow.
fn test_syd_stat_sandbox_stat_allow() -> TestResult {
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdenylist/stat+/dev/null",
            "-mallowlist/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_stat_hide() -> TestResult {
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "-mdenylist/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if stat sandboxing for getdents works to allow.
fn test_syd_stat_sandbox_getdents_allow() -> TestResult {
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdenylist/stat+/dev/null",
            "-mallowlist/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window == b"null"),
        "Stdout:\n{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing for getdents works to hide.
fn test_syd_stat_sandbox_getdents_hide() -> TestResult {
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "-mdenylist/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window != b"null"),
        "Stdout:{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if write sandboxing for open works to allow.
fn test_syd_write_sandbox_open_allow() -> TestResult {
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mdenylist/write+/dev/***",
            "-mallowlist/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if write sandboxing for open works to deny.
fn test_syd_write_sandbox_open_deny() -> TestResult {
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "-mdenylist/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if exec sandboxing works to allow.
fn test_syd_exec_sandbox_open_allow() -> TestResult {
    let bin = which("true")?;
    eprintln!("+ {bin}");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mdenylist/exec+/***",
            &format!("-mallowlist/exec+{bin}"),
            "-Atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if exec sandboxing works to deny.
fn test_syd_exec_sandbox_open_deny() -> TestResult {
    let bin = which("true")?;
    eprintln!("+ {bin}");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mallowlist/exec+/***",
            &format!("-mdenylist/exec+{bin}"),
            "-Atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_ne!(status.code().unwrap_or(127), 0);

    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv4_allow() -> TestResult {
    env::set_var("SYD_DO", "connect4");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+127.0.0.1@4242",
            "-mallowlist/net/connect+127.0.0.1@4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv4_deny() -> TestResult {
    env::set_var("SYD_DO", "connect4");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+127.0.0.1@4242",
            "-mdenylist/net/connect+127.0.0.1@4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 13 /* EACCES */);
    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv6_allow() -> TestResult {
    env::set_var("SYD_DO", "connect6");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+::1@4242",
            "-mallowlist/net/connect+::1@4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv6_deny() -> TestResult {
    env::set_var("SYD_DO", "connect6");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+::1@4242",
            "-mdenylist/net/connect+::1@4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 13 /* EACCES */);
    Ok(())
}

fn test_syd_network_sandbox_allow_successful_bind_ipv4_failure() -> TestResult {
    env::set_var("SYD_DO", "connect4_0");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+127.0.0.1@0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 13 /* EACCES */);
    Ok(())
}

fn test_syd_network_sandbox_allow_successful_bind_ipv4_success() -> TestResult {
    env::set_var("SYD_DO", "connect4_0");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:1",
            "-mallowlist/net/bind+127.0.0.1@0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_network_sandbox_allow_successful_bind_ipv6_failure() -> TestResult {
    env::set_var("SYD_DO", "connect6_0");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/net/bind+::1@0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 13 /* EACCES */);
    Ok(())
}

fn test_syd_network_sandbox_allow_successful_bind_ipv6_success() -> TestResult {
    env::set_var("SYD_DO", "connect6_0");
    let status = syd()
        .args([
            "-mtrace/allow_successful_bind:1",
            "-mallowlist/net/bind+::1@0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_symlink_toctou() -> TestResult {
    env::set_var("SYD_DO", "symlink_toctou");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/read+/etc/passwd",
            "-mdenylist/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_symlinkat_toctou() -> TestResult {
    env::set_var("SYD_DO", "symlinkat_toctou");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/read+/etc/passwd",
            "-mdenylist/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_open() -> TestResult {
    env::set_var("SYD_DO", "ptrmod_toctou_open");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/read+/etc/passwd",
            "-mdenylist/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_creat() -> TestResult {
    env::set_var("SYD_DO", "ptrmod_toctou_creat");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/write+/**/deny.syd-tmp",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_io_uring_escape() -> TestResult {
    env::set_var("SYD_DO", "io_uring_escape");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/read+/etc/passwd",
            "-mdenylist/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_opath_escape() -> TestResult {
    env::set_var("SYD_DO", "opath_escape");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mdenylist/stat+/etc/***",
            "-mallowlist/stat+/etc/ld*",
            "-mdenylist/read+/etc/passwd",
            "-mdenylist/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_1() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_1");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_2() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_2");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_3() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_3");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_4() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_4");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_5() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_5");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_6() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_6");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_7() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_7");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_8() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_8");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_9() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_9");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_10() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_10");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_11() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_11");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_12() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_12");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_13() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_13");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_14() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_14");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_15() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_15");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_16() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_16");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_17() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_17");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_18() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_18");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_19() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_19");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_20() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_chdir_relpath_20");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_1() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_1");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_2() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_2");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_3() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_3");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_4() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_4");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_5() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_5");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_6() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_6");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_7() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_7");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_8() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_8");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_9() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_9");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_10() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_10");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_11() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_11");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_12() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_12");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_13() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_13");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_14() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_14");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_15() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_15");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_16() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_16");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_17() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_17");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_18() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_18");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_19() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_19");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_20() -> TestResult {
    env::set_var("SYD_DO", "devfd_escape_open_relpath_20");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_1() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_1");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_2() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_2");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_3() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_3");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_4() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_4");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_5() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_5");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_6() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_6");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_7() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_7");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_8() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_8");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_9() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_9");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_10() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_10");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_11() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_11");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_12() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_12");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_13() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_13");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_14() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_14");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_15() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_15");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_16() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_16");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_17() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_17");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_18() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_18");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_19() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_19");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_20() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_chdir_relpath_20");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_1() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_1");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_2() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_2");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_3() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_3");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_4() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_4");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_5() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_5");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_6() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_6");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_7() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_7");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_8() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_8");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_9() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_9");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_10() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_10");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_11() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_11");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_12() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_12");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_13() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_13");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_14() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_14");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_15() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_15");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_16() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_16");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_17() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_17");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_18() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_18");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_19() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_19");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_20() -> TestResult {
    env::set_var("SYD_DO", "procself_escape_open_relpath_20");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_relpath() -> TestResult {
    eprintln!("+ grep 'Name:[[:space:]]syd' /proc/./self/status");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            "grep",
            "Name:[[:space:]]syd",
            "/proc/./self/status",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 1,
        "code:{:?}",
        status.code()
    );
    Ok(())
}

fn test_syd_umask_bypass_077() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    env::set_var("SYD_DO", "umask_bypass_077");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_umask_bypass_277() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    env::set_var("SYD_DO", "umask_bypass_277");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_emulate_opath() -> TestResult {
    env::set_var("SYD_DO", "emulate_opath");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallowlist/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_emulate_otmpfile() -> TestResult {
    env::set_var("SYD_DO", "emulate_otmpfile");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_honor_umask() -> TestResult {
    env::set_var("SYD_DO", "honor_umask");
    let prev_umask = umask(Mode::from_bits_truncate(0o077));
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
            "0600",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
            "0644",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    let prev_umask = umask(Mode::from_bits_truncate(0));
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
            "0666",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_open_utf8_invalid() -> TestResult {
    env::set_var("SYD_DO", "open_utf8_invalid");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exec_in_inaccessible_directory() -> TestResult {
    env::set_var("SYD_DO", "exec_in_inaccessible_directory");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-msandbox/write:on",
            "-mallowlist/exec+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_disallow_setuid() -> TestResult {
    let status = syd()
        .args([
            "-ppaludis",
            "-mtrace/allow_unsafe_chmod:0",
            "-msandbox/lock:off",
            "-msandbox/read:off",
            "-msandbox/stat:off",
            "-msandbox/write:off",
            "--",
            "/bin/bash",
            "-c",
            "rm -f test && touch test && chmod u+s test",
        ])
        .status()
        .expect("execute syd");

    assert_ne!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_disallow_setgid() -> TestResult {
    let status = syd()
        .args([
            "-ppaludis",
            "-mtrace/allow_unsafe_chmod:0",
            "-msandbox/lock:off",
            "-msandbox/read:off",
            "-msandbox/stat:off",
            "-msandbox/write:off",
            "--",
            "/bin/bash",
            "-c",
            "rm -f test && touch test && chmod g+s test",
        ])
        .status()
        .expect("execute syd");

    assert_ne!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_pipe() -> TestResult {
    env::set_var("SYD_DO", "fstat_on_pipe");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_socket() -> TestResult {
    env::set_var("SYD_DO", "fstat_on_socket");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallowlist/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_deleted_file() -> TestResult {
    env::set_var("SYD_DO", "fstat_on_deleted_file");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_temp_file() -> TestResult {
    env::set_var("SYD_DO", "fstat_on_temp_file");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_block_ioctl_tiocsti() -> TestResult {
    env::set_var("SYD_DO", "block_ioctl_tiocsti");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_block_prctl_ptrace() -> TestResult {
    env::set_var("SYD_DO", "block_prctl_ptrace");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_block_dev_random() -> TestResult {
    env::set_var("SYD_DO", "block_dev_random");

    let status = syd()
        .args([
            "-ppaludis",
            "-mtrace/allow_unsafe_getrandom:0",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_kill_during_syscall() -> TestResult {
    env::set_var("SYD_DO", "kill_during_syscall");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_open_toolong_path() -> TestResult {
    env::set_var("SYD_DO", "open_toolong_path");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_open_null_path() -> TestResult {
    env::set_var("SYD_DO", "open_null_path");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_utimensat_null() -> TestResult {
    env::set_var("SYD_DO", "utimensat_null");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_normalize_path() -> TestResult {
    const NORMALIZE_PATH_TESTS: [&str; 8] = [
        "null",
        "./null",
        ".////null",
        ".///.////.///./null",
        "./././././././null",
        "./././.././././dev/null",
        "bar/../null",
        "../dev/././../dev/././null",
    ];

    for path in NORMALIZE_PATH_TESTS {
        let status = syd()
            .args([
                "-ppaludis",
                "-msandbox/read:off",
                "-msandbox/stat:off",
                "-msandbox/write:on",
                "-msandbox/exec:off",
                "-mdenylist/write+/***",
                "-mallowlist/write+/dev/null",
                "--",
                "sh",
                "-c",
                &format!("cd /dev; :> {path}"),
            ])
            .status()
            .expect("execute syd");
        assert_eq!(
            status.code().unwrap_or(127),
            0,
            "path:{path}, status:{status:?}"
        );
    }

    Ok(())
}

fn test_syd_path_resolution() -> TestResult {
    env::set_var("SYD_DO", "path_resolution");
    let cwd = syd::fs::canonicalize("/proc/self/cwd", true, syd::fs::MissingHandling::Normal)?
        .to_string_lossy()
        .to_string();

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/",
            "-mallowlist/read+/etc/***",
            "-mallowlist/read+/lib/***",
            "-mallowlist/read+/proc/***",
            "-mallowlist/read+/usr/***",
            "-mallowlist/stat+/etc/***",
            "-mallowlist/stat+/lib/***",
            "-mallowlist/stat+/proc/***",
            "-mallowlist/stat+/usr/***",
            &format!("-mallowlist/read+{cwd}/test_file.txt"),
            &format!("-mallowlist/stat+{cwd}/test_file.txt"),
            &format!("-mallowlist/write+{cwd}/test_file.txt"),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_symlink_readonly_path() -> TestResult {
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "-mdenylist/write+/",
            "--",
            "sh",
            "-c",
            "ln -s / test_syd_symlink_readonly_path && unlink test_syd_symlink_readonly_path",
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_remove_empty_path() -> TestResult {
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            "sh",
            "-c",
            "env LC_ALL=C LANG=C LANGUAGE=C rm '' 2>&1 | tee /dev/stderr | grep -qi 'No such file or directory'"
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_emulate_open_fifo() -> TestResult {
    env::set_var("SYD_DO", "emulate_open_fifo");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_open_trailing_slash() -> TestResult {
    env::set_var("SYD_DO", "open_trailing_slash");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_openat_trailing_slash() -> TestResult {
    env::set_var("SYD_DO", "openat_trailing_slash");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_lstat_trailing_slash() -> TestResult {
    env::set_var("SYD_DO", "lstat_trailing_slash");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fstatat_trailing_slash() -> TestResult {
    env::set_var("SYD_DO", "fstatat_trailing_slash");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_mkdir_trailing_dot() -> TestResult {
    env::set_var("SYD_DO", "mkdir_trailing_dot");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_mkdirat_trailing_dot() -> TestResult {
    env::set_var("SYD_DO", "mkdirat_trailing_dot");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_rmdir_trailing_slashdot() -> TestResult {
    env::set_var("SYD_DO", "rmdir_trailing_slashdot");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fopen_supports_mode_x() -> TestResult {
    env::set_var("SYD_DO", "fopen_supports_mode_x");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_fopen_supports_mode_e() -> TestResult {
    env::set_var("SYD_DO", "fopen_supports_mode_e");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_link_no_symlink_deref() -> TestResult {
    env::set_var("SYD_DO", "link_no_symlink_deref");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_link_posix() -> TestResult {
    env::set_var("SYD_DO", "link_posix");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_linkat_posix() -> TestResult {
    env::set_var("SYD_DO", "linkat_posix");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert_eq!(status.code().unwrap_or(127), 0, "status:{status:?}");
    Ok(())
}

fn test_syd_getcwd_long() -> TestResult {
    env::set_var("SYD_DO", "getcwd_long");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    // FIXME: Fails on musl, possibly not something to fix.
    ignore!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_creat_thru_dangling() -> TestResult {
    env::set_var("SYD_DO", "creat_thru_dangling");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_mkdirat_non_dir_fd() -> TestResult {
    env::set_var("SYD_DO", "mkdirat_non_dir_fd");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_blocking_udp4() -> TestResult {
    env::set_var("SYD_DO", "blocking_udp4");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "-mallowlist/net/bind+LOOPBACK@65432",
            "-mallowlist/net/connect+LOOPBACK@65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_blocking_udp6() -> TestResult {
    env::set_var("SYD_DO", "blocking_udp6");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_successful_bind:0",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "-mallowlist/net/bind+LOOPBACK6@65432",
            "-mallowlist/net/connect+LOOPBACK6@65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_close_on_exec() -> TestResult {
    env::set_var("SYD_DO", "close_on_exec");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_open_exclusive_restart() -> TestResult {
    env::set_var("SYD_DO", "open_exclusive_restart");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    if KERNEL_VERSION.0 > 5 || (KERNEL_VERSION.0 == 5 && KERNEL_VERSION.1 >= 19) {
        eprintln!("Linux kernel version is 5.19 or newer, good.");
        assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    } else {
        eprintln!("Skipping test because the Linux kernel is older than 5.19.");
        ignore!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    }
    Ok(())
}

fn test_syd_open_exclusive_repeat() -> TestResult {
    env::set_var("SYD_DO", "open_exclusive_repeat");

    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_pty_io() -> TestResult {
    env::set_var("SYD_DO", "pty_io");
    if which("gawk").is_err() {
        eprintln!("Test requires gawk, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_diff_dev_fd() -> TestResult {
    env::set_var("SYD_DO", "diff_dev_fd");
    if which("diff").is_err() {
        eprintln!("Test requires diff, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    assert!(status.code().unwrap_or(127) == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_syscall_fuzz() -> TestResult {
    env::set_var("SYD_DO", "syscall_fuzz");
    let epoch = std::time::Instant::now();
    let status = syd()
        .args([
            "-ppaludis",
            "-mlock:off",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/exec:on",
            "-msandbox/net:on",
            "-mallowlist/read+/***",
            "-mallowlist/stat+/***",
            "-mallowlist/exec+/***",
            "-mallowlist/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}
