use std::{
    collections::HashMap,
    env::current_dir,
    ffi::CString,
    fs::File,
    io,
    io::Write,
    os::unix::io::{AsRawFd, RawFd},
    path::{Path, PathBuf},
    ptr,
};

use libseccomp::ScmpFilterContext;
use nix::{
    self,
    errno::Errno,
    fcntl::{fcntl, open, FcntlArg, OFlag},
    libc::{c_char, close},
    sched::clone,
    sys::{
        signal::{kill, SIGCHLD, SIGKILL},
        stat::Mode,
        wait::waitpid,
    },
    unistd::{setpgid, Pid},
};

use crate::unshare::{
    child,
    chroot::{Chroot, Pivot},
    config::Config,
    error::{cmd_result, result, Error, ErrorCode as Err},
    ffi_util::ToCString,
    pipe::{Pipe, PipeHolder},
    stdio::{Closing, Fd},
    Child, Command, ExitStatus,
};

type RawFdPair = (RawFd, RawFd);
type ChildPreExecFunc = Box<dyn Fn() -> Result<(), io::Error>>;

pub struct ChildInfo<'a> {
    pub filename: *const c_char,
    pub args: &'a [*const c_char],
    pub cfg: &'a Config,
    pub chroot: &'a Option<Chroot>,
    pub pivot: &'a Option<Pivot>,
    pub fds: &'a [RawFdPair],
    /// This map may only be used for lookup but not for iteration!
    pub fd_lookup: &'a HashMap<RawFd, RawFd>,
    pub close_fds: &'a [RawFdPair],
    pub pre_exec: &'a Option<ChildPreExecFunc>,
    pub seccomp_filter: &'a Option<ScmpFilterContext>,
    pub seccomp_sender: &'a Option<RawFd>,
    pub host_name: &'a Option<String>,
    pub domain_name: &'a Option<String>,
}

fn raw_with_null(arr: &Vec<CString>) -> Vec<*const c_char> {
    let mut vec = Vec::with_capacity(arr.len().saturating_add(1));
    for i in arr {
        vec.push(i.as_ptr());
    }
    vec.push(ptr::null());
    vec
}

fn relative_to<A: AsRef<Path>, B: AsRef<Path>>(dir: A, rel: B, absolute: bool) -> Option<PathBuf> {
    let dir = dir.as_ref();
    let rel = rel.as_ref();
    let mut dircmp = dir.components();
    for (dc, rc) in rel.components().zip(dircmp.by_ref()) {
        if dc != rc {
            return None;
        }
    }
    if absolute {
        Some(Path::new("/").join(dircmp.as_path()))
    } else {
        Some(dircmp.as_path().to_path_buf())
    }
}

#[allow(clippy::type_complexity)]
fn prepare_descriptors(
    fds: &HashMap<RawFd, Fd>,
) -> Result<
    (
        HashMap<RawFd, RawFd>,
        HashMap<RawFd, PipeHolder>,
        Vec<Closing>,
    ),
    Error,
> {
    let mut inner = HashMap::new();
    let mut outer = HashMap::new();
    let mut guards = Vec::new();
    for (&dest_fd, fdkind) in fds {
        let mut fd = match *fdkind {
            Fd::ReadPipe => {
                let (rd, wr) = Pipe::new()?.split();
                let fd = rd.into_fd();
                guards.push(Closing::new(fd));
                outer.insert(dest_fd, PipeHolder::Writer(wr));
                fd
            }
            Fd::WritePipe => {
                let (rd, wr) = Pipe::new()?.split();
                let fd = wr.into_fd();
                guards.push(Closing::new(fd));
                outer.insert(dest_fd, PipeHolder::Reader(rd));
                fd
            }
            Fd::ReadNull => {
                // Need to keep fd with cloexec, until we are in child
                let fd = result(
                    Err::CreatePipe,
                    open(
                        Path::new("/dev/null"),
                        OFlag::O_CLOEXEC | OFlag::O_RDONLY,
                        Mode::empty(),
                    ),
                )?;
                guards.push(Closing::new(fd));
                fd
            }
            Fd::WriteNull => {
                // Need to keep fd with cloexec, until we are in child
                let fd = result(
                    Err::CreatePipe,
                    open(
                        Path::new("/dev/null"),
                        OFlag::O_CLOEXEC | OFlag::O_WRONLY,
                        Mode::empty(),
                    ),
                )?;
                guards.push(Closing::new(fd));
                fd
            }
            Fd::Inherit => dest_fd,
            Fd::Fd(ref x) => x.as_raw_fd(),
        };
        // The descriptor must not clobber the descriptors that are passed to
        // a child
        while fd != dest_fd && fds.contains_key(&fd) {
            fd = result(Err::CreatePipe, fcntl(fd, FcntlArg::F_DUPFD_CLOEXEC(3)))?;
            guards.push(Closing::new(fd));
        }
        inner.insert(dest_fd, fd);
    }
    Ok((inner, outer, guards))
}

impl Command {
    /// Run the command and return exit status
    pub fn status(&mut self) -> Result<ExitStatus, Error> {
        // TODO(tailhook) stdin/stdout/stderr
        self.spawn()?
            .wait()
            .map_err(|e| Error::WaitError(e.raw_os_error().unwrap_or(-1)))
    }
    /// Spawn the command and return a handle that can be waited for
    pub fn spawn(&mut self) -> Result<Child, Error> {
        // SAFETY: TODO
        unsafe { self.spawn_inner() }
    }

    unsafe fn spawn_inner(&mut self) -> Result<Child, Error> {
        let c_args = raw_with_null(&self.args);

        let (int_fds, ext_fds, _guards) = prepare_descriptors(&self.fds)?;

        let pivot = self
            .pivot_root
            .as_ref()
            .map(|&(ref new, ref old, unmnt)| Pivot {
                new_root: new.to_cstring(),
                put_old: old.to_cstring(),
                #[allow(clippy::disallowed_methods)]
                old_inside: relative_to(old, new, true).unwrap().to_cstring(),
                workdir: current_dir()
                    .ok()
                    .and_then(|cur| relative_to(cur, new, true))
                    .unwrap_or(PathBuf::from("/"))
                    .to_cstring(),
                unmount_old_root: unmnt,
            });

        let chroot = self.chroot_dir.as_ref().map(|dir| {
            let wrk_rel = if let Some((ref piv, _, _)) = self.pivot_root {
                #[allow(clippy::disallowed_methods)]
                piv.join(relative_to(dir, "/", false).unwrap())
            } else {
                dir.to_path_buf()
            };
            Chroot {
                root: dir.to_cstring(),
                workdir: current_dir()
                    .ok()
                    .and_then(|cur| relative_to(cur, wrk_rel, true))
                    .unwrap_or(PathBuf::from("/"))
                    .to_cstring(),
            }
        });

        // Clear the environment.
        std::env::remove_var(crate::config::ENV_LOG);
        std::env::remove_var(crate::config::ENV_NO_SYSLOG);
        std::env::remove_var(crate::config::ENV_NO_CROSS_MEMORY_ATTACH);
        std::env::remove_var(crate::config::ENV_FAKEROOT);
        std::env::remove_var(crate::config::ENV_NPROC);
        std::env::remove_var(crate::config::ENV_SH);
        std::env::remove_var(crate::config::ENV_UNSHARE_MOUNT);
        std::env::remove_var(crate::config::ENV_UNSHARE_UTS);
        std::env::remove_var(crate::config::ENV_UNSHARE_IPC);
        std::env::remove_var(crate::config::ENV_UNSHARE_USER);
        std::env::remove_var(crate::config::ENV_UNSHARE_PID);
        std::env::remove_var(crate::config::ENV_UNSHARE_NET);
        std::env::remove_var(crate::config::ENV_UNSHARE_CGROUP);

        let mut nstack = [0u8; 128 * 1024];
        let args_slice = &c_args[..];
        // We transform all hashmaps into vectors, because iterating over
        // hash map involves closure which crashes in the child in unoptimized
        // build
        let fds = int_fds.iter().map(|(&x, &y)| (x, y)).collect::<Vec<_>>();
        let close_fds = self.close_fds.to_vec();
        let seccomp_sender = self.seccomp_channel.as_ref().map(|c| c.channel());
        let pid = result(
            Err::Fork,
            clone(
                Box::new(|| -> isize {
                    // Note: mo memory allocations/deallocations here
                    let seccomp_sender = if let Some((sender, receiver)) = &seccomp_sender {
                        // SAFETY: In libc, we trust.
                        unsafe { close(receiver.fd) };
                        Some(sender.fd)
                    } else {
                        None
                    };
                    let child_info = ChildInfo {
                        filename: self.filename.as_ptr(),
                        args: args_slice,
                        cfg: &self.config,
                        chroot: &chroot,
                        pivot: &pivot,
                        fds: &fds,
                        fd_lookup: &int_fds,
                        close_fds: &close_fds,
                        pre_exec: &self.pre_exec,
                        seccomp_filter: &self.seccomp_filter,
                        seccomp_sender: &seccomp_sender,
                        host_name: &self.host_name,
                        domain_name: &self.domain_name,
                    };
                    child::child_after_clone(&child_info);
                }),
                &mut nstack[..],
                self.config.namespaces,
                Some(SIGCHLD as i32),
            ),
        )?;

        let seccomp_fd = match self.after_start(pid) {
            Ok(seccomp_fd) => seccomp_fd,
            Err(e) => {
                kill(pid, SIGKILL).ok();
                while let Err(Errno::EINTR) = waitpid(pid, None) {
                    continue;
                }
                return Err(e);
            }
        };

        let mut outer_fds = ext_fds;
        Ok(Child {
            pid: pid.into(),
            status: None,
            stdin: outer_fds.remove(&0).map(|x| match x {
                PipeHolder::Writer(x) => x,
                _ => unreachable!(),
            }),
            stdout: outer_fds.remove(&1).map(|x| match x {
                PipeHolder::Reader(x) => x,
                _ => unreachable!(),
            }),
            stderr: outer_fds.remove(&2).map(|x| match x {
                PipeHolder::Reader(x) => x,
                _ => unreachable!(),
            }),
            fds: outer_fds,
            seccomp_fd,
        })
    }

    #[allow(clippy::cognitive_complexity)]
    fn after_start(&mut self, pid: Pid) -> Result<Option<RawFd>, Error> {
        // Receive the seccomp notification file descriptor.
        let seccomp_fd = if let Some(ref c) = self.seccomp_channel {
            let (sender, receiver) = c.channel();
            // SAFETY: In libc, we trust.
            unsafe { close(sender.fd) };
            match receiver.recvfd() {
                Ok(fd) => {
                    // SAFETY: In libc, we trust.
                    unsafe { close(receiver.fd) };
                    Some(fd)
                }
                Err(Errno::EINVAL) => {
                    // seccomp filter failed to load.
                    errno::set_errno(errno::Errno(Errno::EBUSY as i32));
                    return Err(Errno::EBUSY.into());
                }
                Err(errno) => {
                    return Err(errno.into());
                }
            }
        } else {
            None
        };

        if self.config.make_group_leader {
            result(Err::SetPGid, setpgid(pid, pid))?;
        }

        if let Some((uids, gids)) = self.config.id_maps.as_ref() {
            if let Some((ucmd, gcmd)) = self.id_map_commands.as_ref() {
                let mut cmd = Command::new(ucmd);
                cmd.arg(format!("{}", pid));
                for map in uids {
                    cmd.arg(format!("{}", map.inside_uid));
                    cmd.arg(format!("{}", map.outside_uid));
                    cmd.arg(format!("{}", map.count));
                }
                cmd_result(Err::SetIdMap, cmd.status())?;
                let mut cmd = Command::new(gcmd);
                cmd.arg(format!("{}", pid));
                for map in gids {
                    cmd.arg(format!("{}", map.inside_gid));
                    cmd.arg(format!("{}", map.outside_gid));
                    cmd.arg(format!("{}", map.count));
                }
                cmd_result(Err::SetIdMap, cmd.status())?;
            } else {
                {
                    // Write "deny" to /proc/[pid]/setgroups before writing to gid_map.
                    let path = format!("/proc/{pid}/setgroups");
                    if File::create(path)
                        .and_then(|mut f| f.write_all(b"deny"))
                        .is_err()
                    {
                        return Err(Error::SetIdMap(Errno::last() as i32));
                    }
                }
                let mut buf = Vec::new();
                for map in gids {
                    #[allow(clippy::disallowed_methods)]
                    if writeln!(
                        &mut buf,
                        "{} {} {}",
                        map.inside_gid, map.outside_gid, map.count
                    )
                    .is_err()
                    {
                        return Err(Error::SetIdMap(Errno::last() as i32));
                    }
                }
                if File::create(format!("/proc/{pid}/gid_map"))
                    .and_then(|mut f| f.write_all(&buf[..]))
                    .is_err()
                {
                    return Err(Error::SetIdMap(Errno::last() as i32));
                }
                let mut buf = Vec::new();
                for map in uids {
                    #[allow(clippy::disallowed_methods)]
                    if writeln!(
                        &mut buf,
                        "{} {} {}",
                        map.inside_uid, map.outside_uid, map.count
                    )
                    .is_err()
                    {
                        return Err(Error::SetIdMap(Errno::last() as i32));
                    }
                }
                if File::create(format!("/proc/{pid}/uid_map"))
                    .and_then(|mut f| f.write_all(&buf[..]))
                    .is_err()
                {
                    return Err(Error::SetIdMap(Errno::last() as i32));
                }
            }
        }

        if let Some(ref mut callback) = self.before_unfreeze {
            #[allow(clippy::cast_sign_loss)]
            callback(i32::from(pid) as u32).map_err(Error::BeforeUnfreeze)?;
        }

        Ok(seccomp_fd)
    }
}
